import logging
import string
from typing import *
from collections import Counter

import numpy as np
import openpyxl
from openpyxl.styles import Alignment, Font, NamedStyle, PatternFill
from openpyxl.worksheet.worksheet import Worksheet
from sklearn.cluster import KMeans
from sklearn.decomposition import PCA

from .. import QGRAIN_VERSION
from ..models import DistributionType, Dataset, Sample, ArtificialDataset, SSUResult, EMMAResult , UDMResult
from ..statistics import _all_scales, all_statistics, logarithmic
from ..distributions import get_distribution


def column_to_char(column_index: int):
    column = column_index + 1
    column_str = str()
    while column != 0:
        res = column % 26
        if res == 0:
            res = 26
            column -= 26
        column_str = chr(ord('A') + res - 1) + column_str
        column = column // 26
    return column_str


def to_cell_name(row: int, column: int):
    return "{0}{1}".format(column_to_char(column), row+1)


def prepare_styles(wb: openpyxl.Workbook):
    normal_light_style = NamedStyle(name="normal_light")
    normal_light_style.font = Font(size=12, name="Arial", color="000000")
    normal_light_style.alignment = Alignment(
        horizontal='center',
        vertical='center',
        text_rotation=0,
        wrap_text=False,
        shrink_to_fit=False,
        indent=0)
    normal_light_style.fill = PatternFill(
        patternType="solid",
        fgColor="c4cbcf")
    wb.add_named_style(normal_light_style)

    normal_dark_style = NamedStyle(name="normal_dark")
    normal_dark_style.font = Font(size=12, name="Arial", color="000000")
    normal_dark_style.alignment = Alignment(
        horizontal='center',
        vertical='center',
        text_rotation=0,
        wrap_text=False,
        shrink_to_fit=False,
        indent=0)
    normal_dark_style.fill = PatternFill(
        patternType="solid",
        fgColor="b2bbbe")
    wb.add_named_style(normal_dark_style)

    warning_style = NamedStyle(name="warning")
    warning_style.font = Font(size=12, name="Arial", color="ffffff")
    warning_style.alignment = Alignment(
        horizontal='center',
        vertical='center',
        text_rotation=0,
        wrap_text=False,
        shrink_to_fit=False,
        indent=0)
    warning_style.fill = PatternFill(
        patternType="solid",
        fgColor="ff9900")
    wb.add_named_style(warning_style)

    header_style = NamedStyle(name="header")
    header_style.font = Font(size=14, bold=True, name="Times New Roman", color="ffffff")
    header_style.alignment = Alignment(
        horizontal='center',
        vertical='center',
        text_rotation=0,
        wrap_text=True,
        shrink_to_fit=False,
        indent=0)
    header_style.fill = PatternFill(
        patternType="solid",
        fgColor="475164")
    wb.add_named_style(header_style)

    description_style = NamedStyle(name="description")
    description_style.font = Font(size=14, bold=True, name="Times New Roman", color="ffffff")
    description_style.alignment = Alignment(
        horizontal='left',
        vertical='center',
        text_rotation=0,
        wrap_text=False,
        shrink_to_fit=False,
        indent=0)
    description_style.fill = PatternFill(
        patternType="solid",
        fgColor="475164")
    wb.add_named_style(description_style)


SMALL_WIDTH = 12
MEDIAN_WIDTH = 24
LARGE_WIDTH = 48


def _write_readme_sheet(ws: Worksheet, text: str):
    ws.title = "README"
    full_text = \
        """
        This Excel file was generated by QGrain ({0}).

        Please cite:
        Liu, Y., Liu, X., Sun, Y., 2021. QGrain: An open-source and easy-to-use software for the comprehensive
            analysis of grain size distributions. Sedimentary Geology 423, 105980. DOI: 10.1016/j.sedgeo.2021.105980

        {1}

        """.format(QGRAIN_VERSION, text)
    lines_of_description = full_text.split("\n")
    for row, line in enumerate(lines_of_description):
        cell = ws.cell(row+1, 1, value=line)
        cell.style = "description"
    ws.column_dimensions[column_to_char(0)].width = 200


def _write_dataset_sheet(ws: Worksheet, dataset: Dataset, progress_callback: Callable[[float], None] = None):
    def write(r, c, v, s="normal_light"):
        cell = ws.cell(r+1, c+1, value=v)
        cell.style = s

    ws.title = "GSDs"
    write(0, 0, "Sample Name", s="header")
    ws.column_dimensions[column_to_char(0)].width = 24
    for col, value in enumerate(dataset.classes, 1):
        write(0, col, value, s="header")
        ws.column_dimensions[column_to_char(col)].width = 10

    for i_sample, sample in enumerate(dataset):
        row = i_sample + 1
        if row % 2 == 0:
            style = "normal_dark"
        else:
            style = "normal_light"
        write(row, 0, sample.name, s=style)
        for col, value in enumerate(sample.distribution, 1):
            write(row, col, value, s=style)
        if progress_callback is not None:
            progress = i_sample / len(dataset)
            progress_callback(progress)


def save_artificial_dataset(dataset: ArtificialDataset, filename: str,
                            progress_callback: Callable[[float], None] = None,
                            logger: logging.Logger = None):
    if logger is None:
        logger = logging.getLogger("QGrain")
    else:
        assert isinstance(logger, logging.Logger)
    logger.debug("Start to save the artificial dataset.")
    wb = openpyxl.Workbook()
    prepare_styles(wb)
    logger.debug("Creating the README sheet.")
    readme_text = \
        f"""
        It contains {dataset.n_components+2} sheets:
        1. The first sheet is the grain size distributions of this artificial dataset.
        2. The second sheet stores the parameters which were used to generate the dataset.
        3-{dataset.n_components+2}. Each of the left sheets is used to put the distributions of each component.

        Sampling Settings
            Minimum size [micron]: {dataset.classes[0]}
            Maximum size [micron]: {dataset.classes[-1]}
            Number of Grain Size Classes: {len(dataset.classes)}
            Precision: {dataset.precision}
            Noise Decimals: {dataset.noise}
            Number of Samples: {len(dataset)}
        """
    _write_readme_sheet(wb.active, readme_text)

    def write(r, c, v, s="normal_light"):
        cell = ws.cell(r + 1, c + 1, value=v)
        cell.style = s

    distribution_class = get_distribution(dataset.distribution_type)
    parameter_names = list(distribution_class.PARAMETER_NAMES) + ["Weight"]

    logger.debug("Creating the GSDs sheet.")
    ws = wb.create_sheet("GSDs")
    write(0, 0, "Sample Name", s="header")
    ws.column_dimensions[column_to_char(0)].width = 24
    for col, value in enumerate(dataset.classes, 1):
        write(0, col, value, s="header")
        ws.column_dimensions[column_to_char(col)].width = 10
    for row, sample in enumerate(dataset, 1):
        if row % 2 == 0:
            style = "normal_dark"
        else:
            style = "normal_light"
        write(row, 0, sample.name, s=style)
        for col, value in enumerate(sample.distribution, 1):
            write(row, col, value, s=style)
        if progress_callback is not None:
            progress_callback(row / len(dataset) * 0.2)

    logger.debug("Creating the Parameters sheet.")
    ws = wb.create_sheet("Parameters")
    write(0, 0, "Sample Name", s="header")
    ws.merge_cells(start_row=1, start_column=1, end_row=2, end_column=1)
    ws.column_dimensions[column_to_char(0)].width = 24
    for i in range(dataset.n_components):
        write(0, dataset.n_parameters * i + 1, f"Component{i+1}", s="header")
        ws.merge_cells(start_row=1, start_column=dataset.n_parameters*i+2,
                       end_row=1, end_column=dataset.n_parameters*(i+1)+1)
        for j, header_name in enumerate(parameter_names):
            write(1, dataset.n_parameters * i + 1 + j, header_name, s="header")
            ws.column_dimensions[column_to_char(dataset.n_parameters*i+1+j)].width = 16
    for i, sample in enumerate(dataset):
        row = i + 2
        if row % 2 == 1:
            style = "normal_dark"
        else:
            style = "normal_light"
        write(row, 0, sample.name, s=style)
        for j, component in enumerate(sample):
            for k in range(dataset.n_parameters):
                write(row, dataset.n_parameters * j + k + 1, dataset.parameters[i, k, j], s=style)
        if progress_callback is not None:
            progress_callback(1 / len(dataset) * 0.2 + 0.2)

    for i in range(dataset.n_components):
        logger.debug(f"Creating the C{i+1} sheet.")
        ws = wb.create_sheet(f"C{i+1}")
        write(0, 0, "Sample Name", s="header")
        ws.column_dimensions[column_to_char(0)].width = 24
        for col, value in enumerate(dataset.classes, 1):
            write(0, col, value, s="header")
            ws.column_dimensions[column_to_char(col)].width = 10
        for row, sample in enumerate(dataset, 1):
            if row % 2 == 0:
                style = "normal_dark"
            else:
                style = "normal_light"
            write(row, 0, sample.name, s=style)
            for col, value in enumerate(sample[i].distribution, 1):
                write(row, col, value, s=style)
            if progress_callback is not None:
                progress_callback(((i*len(dataset) + row) / len(dataset)*dataset.n_components) * 0.6 + 0.4)

    wb.save(filename)
    wb.close()
    if progress_callback is not None:
        progress_callback(1.0)
    logger.info(f"The artificial dataset has been saved to the Excel file: [{filename}].")


def save_dataset(dataset: Dataset, filename: str, progress_callback: Callable[[float], None] = None,
                 logger: logging.Logger = None):
    assert dataset is not None
    if logger is None:
        logger = logging.getLogger("QGrain")
    else:
        assert isinstance(logger, logging.Logger)
    logger.debug("Start to save the grain size dataset.")
    wb = openpyxl.Workbook()
    prepare_styles(wb)
    logger.debug("Creating the README sheet.")
    readme_text = \
        """
        It only contains one sheet which stores the grain size distributions.
        """
    _write_readme_sheet(wb.active, readme_text)
    logger.debug("Creating the GSDs sheet.")
    ws = wb.create_sheet("GSDs")
    _write_dataset_sheet(ws, dataset, progress_callback)
    logger.debug("Saving the workbook to file.")
    wb.save(filename)
    wb.close()
    if progress_callback is not None:
        progress_callback(1.0)
    logger.info(f"The dataset has been saved to the Excel file: [{filename}].")


def save_statistics(dataset: Dataset, filename: str, progress_callback: Callable[[float], None] = None,
                    logger: logging.Logger = None):
    assert dataset is not None
    if logger is None:
        logger = logging.getLogger("QGrain")
    else:
        assert isinstance(logger, logging.Logger)
    logger.debug("Start to save the statistical result.")
    # Calculate
    logger.debug("Calculating the statistical parameters and classification groups of all samples.")
    all_sample_statistics = []
    for i, sample in enumerate(dataset):
        sample_statistics = all_statistics(sample.classes, sample.classes_phi, sample.distribution)
        all_sample_statistics.append(sample_statistics)
        if progress_callback is not None:
            progress = (i / len(dataset)) * 0.4
            progress_callback(progress)
    wb = openpyxl.Workbook()
    prepare_styles(wb)
    logger.debug("Creating the README sheet.")
    readme_text = \
        """
        It contains 6 sheets.
            1-5. The previous five sheets stores the statistical parameters of five computational methods, respectively.
            6. The last sheet puts the proportions of different size scales and the classification groups.

        The statistical formulas are referred to Blott & Pye (2001)'s work.
        The classification of GSDs is referred to Folk (1954)'s and Blott & Pye (2012)'s scheme.

        References:
            1. Blott, S. J. & Pye, K. Particle size scales and classification of sediment types based on particle size
                distributions: Review and recommended procedures. Sedimentology 59, 2071–2096 (2012).
            2. Blott, S. J. & Pye, K. GRADISTAT: a grain-size distribution and statistics package for the analysis of
                unconsolidated sediments. Earth Surf. Process. Landforms 26, 1237–1248 (2001).
            3. Folk, R. L. The Distinction between Grain Size and Mineral Composition in Sedimentary-Rock Nomenclature.
                The Journal of Geology 62, 344–359 (1954).
        """
    _write_readme_sheet(wb.active, readme_text)
    logger.debug("Creating the GSDs sheet.")
    if progress_callback is not None:
        _callback = lambda p: progress_callback(p * 0.1 + 0.4)
    else:
        _callback = None
    ws = wb.create_sheet("GSDs")
    _write_dataset_sheet(ws, dataset, progress_callback=_callback)

    def get_keys(method: str):
        small_width = 12
        median_width = 24
        large_width = 48
        if method == "arithmetic":
            unit = "μm"
            keys = [
                (lambda s: s[method]["mean"], f"Mean [{unit}]", small_width),
                (lambda s: s[method]["std"], f"Sorting Coefficient", small_width),
                (lambda s: s[method]["skewness"], f"Skewness", small_width),
                (lambda s: s[method]["kurtosis"], f"Kurtosis", small_width)]
            return keys
        elif method in ("geometric", "logarithmic", "geometric_fw57", "logarithmic_fw57"):
            unit = "μm" if method.startswith("geometric") else "φ"
            keys = [
                (lambda s: s[method]["mean"], f"Mean [{unit}]", small_width),
                (lambda s: s[method]["mean_description"], "Mean Description", median_width),
                (lambda s: s[method]["first_percentile"], f"First Percentile (C) [{unit}]", small_width),
                (lambda s: s[method]["median"], f"Median (M) [{unit}]", small_width),
                (lambda s: s[method]["mode"], f"Mode [{unit}]", small_width),
                (lambda s: len(s[method]["modes"]), f"Number of Modes", small_width),
                (lambda s: ", ".join([f"{mode: 0.4f}" for mode in s[method]["modes"]]), f"Modes [{unit}]", median_width),
                (lambda s: s[method]["proportion_ss"] * 100.0, "Sortable Silt Proportion [%]", small_width),
                (lambda s: s[method]["mean_ss"], f"Mean Sortable Silt Size [{unit}]", small_width),
                (lambda s: s[method]["std"], "Sorting Coefficient", small_width),
                (lambda s: s[method]["std_description"], "Sorting Description", median_width),
                (lambda s: s[method]["skewness"], "Skewness", small_width),
                (lambda s: s[method]["skewness_description"], "Skewness Description", median_width),
                (lambda s: s[method]["kurtosis"], "Kurtosis", small_width),
                (lambda s: s[method]["kurtosis_description"], "Kurtosis Description", median_width)]
            return keys
        elif method == "proportion_and_classification":
            keys = [
                (lambda s: ", ".join([f"{p*100:0.4f}" for p in s["proportions_gsm"]]), "(Gravel, Sand, Mud) Proportions [%]", large_width),
                (lambda s: ", ".join([f"{p*100:0.4f}" for p in s["proportions_ssc"]]), "(Sand, Silt, Clay) Proportions [%]", large_width),
                (lambda s: ", ".join([f"{p*100:0.4f}" for p in s["proportions_bgssc"]]), "(Boulder, Gravel, Sand, Silt, Clay) Proportions [%]", large_width),
                (lambda s: s["group_folk54"], "Group (Folk, 1954)", median_width),
                (lambda s: s["group_bp12"], "Group (Blott & Pye, 2012)", large_width),
                (lambda s: s["group_bp12_symbol"], "Group Symbol (Blott & Pye, 2012)", median_width)]
            all_scales = _all_scales()
            for scale in all_scales:
                func = lambda s, scale=scale: s["proportions"][scale] * 100.0
                name = string.capwords(" ".join(scale)) + " Proportion [%]"
                keys.append((func, name, small_width))
            return keys
        else:
            raise NotImplementedError(method)

    def write(r, c, v, s="normal_light"):
        cell = ws.cell(r + 1, c + 1, value=v)
        cell.style = s

    methods = ["arithmetic", "geometric", "logarithmic", "geometric_fw57",
               "logarithmic_fw57", "proportion_and_classification"]
    sheet_names = ["Arithmetic", "Geometric", "Logarithmic", "Geometric_fw57",
                   "Logarithmic_fw57", "Proportion and Classification"]
    for i_method, (method, sheet_name) in enumerate(zip(methods, sheet_names)):
        logger.debug(f"Creating the {sheet_name} sheet.")
        ws = wb.create_sheet(sheet_name)
        write(0, 0, "Sample Name", s="header")
        keys = get_keys(method)
        ws.column_dimensions[column_to_char(0)].width = 16
        for col, (func, name, width) in enumerate(keys, 1):
            write(0, col, name, s="header")
            ws.column_dimensions[column_to_char(col)].width = width
        for i_sample, (sample, sample_statistics) in enumerate(zip(dataset, all_sample_statistics)):
            row = i_sample + 1
            if row % 2 == 0:
                style = "normal_dark"
            else:
                style = "normal_light"
            write(row, 0, sample.name, s=style)
            for col, (func, name, width) in enumerate(keys, 1):
                value = func(sample_statistics)
                write(row, col, value, s=style)
            if progress_callback is not None:
                progress = 0.5 + ((i_sample / len(dataset)) + i_method) / len(methods) * 0.5
                progress_callback(progress)
    logger.debug("Saving the workbook to file.")
    wb.save(filename)
    wb.close()
    if progress_callback is not None:
        progress_callback(1.0)
    logger.info(f"The statistical result has been saved to the Excel file: [{filename}].")


def save_pca(dataset: Dataset, filename: str, progress_callback: Callable[[float], None] = None,
             logger: logging.Logger = None):
    assert dataset is not None
    if logger is None:
        logger = logging.getLogger("QGrain")
    else:
        assert isinstance(logger, logging.Logger)
    logger.debug("Start to save the PCA result.")
    # Calculate
    logger.debug("Performing the PCA algorithm to this dataset. The number of PCs is set to 10.")
    pca = PCA(n_components=10)
    transformed = pca.fit_transform(dataset.distributions)
    components = pca.components_
    ratios = pca.explained_variance_ratio_

    wb = openpyxl.Workbook()
    prepare_styles(wb)
    logger.debug("Creating the README sheet.")
    readme_text = \
        """
        It contains 3 sheets:
        1. The first sheet is the dataset which was used to perform the PCA algorithm.
        2. The second sheet is used to put the distributions of all PCs.
        3. The third sheet is used to store the PC variations of all samples.

        The base PCA algorithm is implemented by scikit-learn.
        You can get the details of algorithm from the following website.
        https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.PCA.html
        """
    _write_readme_sheet(wb.active, readme_text)
    logger.debug("Creating the GSDs sheet.")
    if progress_callback is not None:
        _callback = lambda progress: progress_callback(progress * 0.2)
    else:
        _callback = None
    ws = wb.create_sheet("GSDs")
    _write_dataset_sheet(ws, dataset, progress_callback=_callback)

    def write(r, c, v, s="normal_light"):
        cell = ws.cell(r + 1, c + 1, value=v)
        cell.style = s

    logger.debug("Creating the Distributions of PCs sheet.")
    ws = wb.create_sheet("Distributions of PCs")
    write(0, 0, "PC", s="header")
    ws.column_dimensions[column_to_char(0)].width = 16
    for col, value in enumerate(dataset.classes, 1):
        write(0, col, value, s="header")
        ws.column_dimensions[column_to_char(col)].width = 10
    for i, component in enumerate(components):
        row = i + 1
        if row % 2 == 0:
            style = "normal_dark"
        else:
            style = "normal_light"
        write(row, 0, f"PC{i+1} ({ratios[i]:0.2%})", s=style)
        for col, value in enumerate(component, 1):
            write(row, col, value, s=style)
    if progress_callback is not None:
        progress_callback(0.3)

    logger.debug("Creating the Variations of PCs sheet.")
    ws = wb.create_sheet("Variations of PCs")
    write(0, 0, "Sample Name", s="header")
    ws.column_dimensions[column_to_char(0)].width = 16
    for i in range(10):
        write(0, i + 1, f"PC{i+1} ({ratios[i]:0.2%})", s="header")
        ws.column_dimensions[column_to_char(i+1)].width = 10
    for row, variations in enumerate(transformed, 1):
        if row % 2 == 0:
            style = "normal_dark"
        else:
            style = "normal_light"
        write(row, 0, dataset[row-1].name, s=style)
        for col, value in enumerate(variations, 1):
            write(row, col, value, s=style)
        if progress_callback is not None:
            progress_callback(row / len(dataset) * 0.7 + 0.3)

    wb.save(filename)
    wb.close()
    if progress_callback is not None:
        progress_callback(1.0)
    logger.info(f"The PCA result has been saved to the Excel file: [{filename}].")


def save_clustering(dataset: Dataset, flags: Sequence[int], filename: str,
                    data_key: Tuple, linkage_method: str, metric: str,
                    progress_callback: Callable[[float], None] = None, logger: logging.Logger = None):
    assert dataset is not None
    if logger is None:
        logger = logging.getLogger("QGrain")
    else:
        assert isinstance(logger, logging.Logger)
    logger.debug("Start to save the clustering result.")
    flag_set = set(flags)
    n_clusters = len(flag_set)
    typical_samples: List[Sample] = []
    temp_flag_set = set()
    for i, flag in enumerate(flags):
        if len(temp_flag_set) == n_clusters:
            break
        if flag not in temp_flag_set:
            typical_samples.append(dataset[i])
            temp_flag_set.add(flag)

    wb = openpyxl.Workbook()
    prepare_styles(wb)
    logger.debug("Creating the README sheet.")
    readme_text = \
        f"""
        It contains 3 or {n_clusters+3} sheets:
        1. The first sheet is the dataset which was used to perform the hierarchy clustering algorithm.
        2. The second sheet is used to put the clustering flags of all samples.
        3. The third sheet is the typical samples (i.e, the first sample of each cluster was selected).
        4. If the number of clusters less equal to 100, the samples of each cluster will be save to individual sheets.

        The base hierarchy clustering algorithm is implemented by Scipy.
        You can get the details of algorithm from the following website.
        https://docs.scipy.org/doc/scipy/reference/cluster.hierarchy.html
        
        Data key: {data_key}
        Linkage method: {linkage_method}
        Metric: {metric}
        """
    _write_readme_sheet(wb.active, readme_text)
    logger.debug("Creating the GSDs sheet.")
    if progress_callback is not None:
        _callback = lambda p: progress_callback(p * 0.2)
    else:
        _callback = None
    ws = wb.create_sheet("GSDs")
    _write_dataset_sheet(ws, dataset, progress_callback=_callback)

    def write(r, c, v, s="normal_light"):
        cell = ws.cell(r + 1, c + 1, value=v)
        cell.style = s

    logger.debug("Creating the Cluster Flags of Samples sheet.")
    ws = wb.create_sheet("Cluster Flags of Samples")
    write(0, 0, "Sample Name", s="header")
    write(0, 1, "Cluster Flags", s="header")
    ws.column_dimensions[column_to_char(0)].width = 16
    ws.column_dimensions[column_to_char(1)].width = 16
    for i, (sample, flag) in enumerate(zip(dataset, flags)):
        row = i + 1
        if row % 2 == 0:
            style = "normal_dark"
        else:
            style = "normal_light"
        write(row, 0, sample.name, s=style)
        write(row, 1, flag, s=style)
        if progress_callback is not None:
            if n_clusters <= 100:
                progress_callback(i / len(dataset) * 0.1 + 0.2)
            else:
                progress_callback(i / len(dataset) * 0.4 + 0.2)

    logger.debug("Creating the Typical Samples of Clusters sheet.")
    ws = wb.create_sheet("Typical Samples of Clusters")
    write(0, 0, "Sample Name", s="header")
    ws.column_dimensions[column_to_char(0)].width = 16
    for col, value in enumerate(dataset.classes, 1):
        write(0, col, value, s="header")
        ws.column_dimensions[column_to_char(col)].width = 10
    for i, sample in enumerate(typical_samples):
        row = i + 1
        if row % 2 == 0:
            style = "normal_dark"
        else:
            style = "normal_light"
        write(row, 0, sample.name, s=style)
        for col, value in enumerate(sample.distribution, 1):
            write(row, col, value, s=style)
        if progress_callback is not None:
            if n_clusters <= 100:
                progress_callback(i / len(dataset) * 0.1 + 0.3)
            else:
                progress_callback(i / len(dataset) * 0.4 + 0.6)

    if n_clusters <= 100:
        for flag in flag_set:
            samples = []
            for sample, in_this_cluster in zip(dataset, np.equal(flags, flag)):
                if in_this_cluster:
                    samples.append(sample)
            logger.debug(f"Creating the Cluster{flag} sheet.")
            ws = wb.create_sheet(f"Cluster{flag}")
            write(0, 0, "Sample Name", s="header")
            ws.column_dimensions[column_to_char(0)].width = 16
            for col, value in enumerate(dataset.classes, 1):
                write(0, col, value, s="header")
                ws.column_dimensions[column_to_char(col)].width = 10
            for i, sample in enumerate(samples):
                row = i + 1
                if row % 2 == 0:
                    style = "normal_dark"
                else:
                    style = "normal_light"
                write(row, 0, sample.name, s=style)
                for col, value in enumerate(sample.distribution, 1):
                    write(row, col, value, s=style)
                if progress_callback is not None:
                    progress_callback(i / len(dataset) / n_clusters * 0.6 + 0.4)

    wb.save(filename)
    wb.close()
    if progress_callback is not None:
        progress_callback(1.0)
    logger.info(f"The Clustering result has been saved to the Excel file: [{filename}].")


def save_emma(result: EMMAResult, filename: str, progress_callback: Callable[[float], None] = None,
              logger: logging.Logger = None):
    if logger is None:
        logger = logging.getLogger("QGrain")
    else:
        assert isinstance(logger, logging.Logger)
    logger.debug("Start to save the EMMA result.")

    wb = openpyxl.Workbook()
    prepare_styles(wb)
    logger.debug("Creating the README sheet.")
    readme_text = \
        f"""
        It contains three sheets:
        1. The first sheet is the dataset which was used to perform the EMMA algorithm.
        2. The second sheet is used to put the distributions of all end members.
        3. The third sheet is the proportions of end members of all samples.

        This EMMA algorithm was implemented by QGrain, using the famous machine learning framework, PyTorch.

        EMMA Algorithm Details
            Number of Samples: {result.n_samples}
            Kernel Type: {result.kernel_type.name}
            Number of End Members: {result.n_members}
            x0: {result.x0 if result.x0 is None else result.x0.tolist()}
            Number of Epochs: {len(result.loss_series("lmse"))}
            Spent Time: {result.time_spent} s

            Computing Device: {"Unknown" if result.settings is None else result.settings["device"]}
            Loss Function: {"Unknown" if result.settings is None else result.settings["loss"]}
            Number of Pretrain Epochs: {"Unknown" if result.settings is None else result.settings["pretrain_epochs"]}
            Minimum Number of Epochs: {"Unknown" if result.settings is None else result.settings["min_epochs"]}
            Maximum Number of Epochs: {"Unknown" if result.settings is None else result.settings["max_epochs"]}
            Learning Rate: {"Unknown" if result.settings is None else result.settings["learning_rate"]}
            Precision: {"Unknown" if result.settings is None else result.settings["precision"]}
            Betas: {"Unknown" if result.settings is None else result.settings["betas"]}
            Update End Members: {"Unknown" if result.settings is None else result.settings["update_end_members"]}
            Need History: {"Unknown" if result.settings is None else result.settings["need_history"]}
        """
    _write_readme_sheet(wb.active, readme_text)
    logger.debug("Creating the GSDs sheet.")
    if progress_callback is not None:
        _callback = lambda p: progress_callback(p * 0.4)
    else:
        _callback = None
    ws = wb.create_sheet("GSDs")
    _write_dataset_sheet(ws, result.dataset, progress_callback=_callback)

    def write(r, c, v, s="normal_light"):
        cell = ws.cell(r + 1, c + 1, value=v)
        cell.style = s

    logger.debug("Creating the Distributions of End Members sheet.")
    ws = wb.create_sheet("Distributions of End Members")
    write(0, 0, "End Member", s="header")
    ws.column_dimensions[column_to_char(0)].width = 16
    for col, value in enumerate(result.dataset.classes, 1):
        write(0, col, value, s="header")
        ws.column_dimensions[column_to_char(col)].width = 10
    for i in range(result.n_members):
        row = i + 1
        if row % 2 == 0:
            style = "normal_dark"
        else:
            style = "normal_light"
        write(row, 0, f"EM{i+1}", s=style)
        for col, value in enumerate(result.end_members[i], 1):
            write(row, col, value, s=style)

    logger.debug("Creating the Proportions of End Members sheet.")
    ws = wb.create_sheet("Proportions of End Members")
    write(0, 0, "Sample Name", s="header")
    ws.column_dimensions[column_to_char(0)].width = 16
    for i in range(result.n_members):
        write(0, i + 1, f"EM{i+1} [%]", s="header")
        ws.column_dimensions[column_to_char(i+1)].width = 10
    for i, sample_proportions in enumerate(result.proportions):
        row = i + 1
        if row % 2 == 0:
            style = "normal_dark"
        else:
            style = "normal_light"
        write(row, 0, result.dataset[i].name, s=style)
        for j in range(result.n_members):
            col = j + 1
            write(row, col, sample_proportions[j]*100, s=style)
        if progress_callback is not None:
            progress_callback(i / result.n_samples * 0.6 + 0.4)

    wb.save(filename)
    wb.close()
    if progress_callback is not None:
        progress_callback(1.0)
    logger.info(f"The EMMA result has been saved to the Excel file: [{filename}].")


def save_ssu(results: List[SSUResult], filename: str, align_components=False,
             progress_callback: Callable[[float], None] = None, logger: logging.Logger = None):
    if logger is None:
        logger = logging.getLogger("QGrain")
    else:
        assert isinstance(logger, logging.Logger)
    logger.debug("Start to save the SSU results.")
    # pack the GSDs of samples to the dataset
    dataset = Dataset(
        "GSDs", [result.sample.name for result in results],
        results[0].classes,
        [result.sample.distribution for result in results])
    max_n_components = max(Counter([len(result) for result in results]).keys())

    # prepare flags
    flags = []
    if not align_components:
        for result in results:
            flags.extend(range(len(result)))
    else:
        stacked_components = []
        for result in results:
            for component in result:
                stacked_components.append(component.distribution)
        stacked_components = np.array(stacked_components)
        clusters = KMeans(n_clusters=max_n_components)
        flags = clusters.fit_predict(stacked_components)
        # check flags to make it unique
        flag_index = 0
        for result in results:
            result_flags = set()
            for i in range(len(result)):
                if flags[flag_index] in result_flags:
                    if flags[flag_index] == max_n_components:
                        flags[flag_index] = max_n_components-1
                    else:
                        flags[flag_index] += 1
                    result_flags.add(flags[flag_index])
                flag_index += 1

        flag_set = set(flags)
        picked = []
        for target_flag in flag_set:
            for i, flag in enumerate(flags):
                if flag == target_flag:
                    picked.append((target_flag, logarithmic(dataset.classes_phi, stacked_components[i])["mean"]))
                    break
        picked.sort(key=lambda x: x[1])
        flag_map = {flag: index for index, (flag, _) in enumerate(picked)}
        flags = np.array([flag_map[flag] for flag in flags])

    wb = openpyxl.Workbook()
    prepare_styles(wb)
    logger.debug("Creating the README sheet.")
    readme_text = \
        f"""
        It contains {max_n_components+4} sheets:
        1. The first sheet is used to put the grain size distributions of corresponding samples.
        2. The second sheet is used to put the fitting information and resolved parameters of SSU results.
        3. The third sheet stores the statistical parameters of all components.
        4. The fourth sheet is used to put the distributions of unmixed components and the sum.
        5-{max_n_components+4}. Other sheets severally store the distributions of each component group.

        The SSU algorithm is implemented by QGrain.
        """
    _write_readme_sheet(wb.active, readme_text)
    logger.debug("Creating the GSDs sheet.")
    if progress_callback is not None:
        _callback = lambda p: progress_callback(p * 0.1)
    else:
        _callback = None
    ws = wb.create_sheet("GSDs")
    _write_dataset_sheet(ws, dataset, progress_callback=_callback)

    def write(r, c, v, s="normal_light"):
        cell = ws.cell(r + 1, c + 1, value=v)
        cell.style = s

    logger.debug("Creating the Information of Fitting sheet.")
    ws = wb.create_sheet("Information of Fitting")
    write(0, 0, "Sample Name", s="header")
    ws.column_dimensions[column_to_char(0)].width = 16
    headers = ["Distribution Type",
               "Number of Components",
               "Resolver Settings",
               "x0",
               "Resolved Parameters",
               "Spent Time [s]",
               "Number of Iterations",
               "Final Loss [LMSE]"]
    for col, value in enumerate(headers, 1):
        write(0, col, value, s="header")
        if col in (3, 4, 5):
            ws.column_dimensions[column_to_char(col)].width = 32
        else:
            ws.column_dimensions[column_to_char(col)].width = 10
    for i, result in enumerate(results):
        row = i + 1
        if row % 2 == 0:
            style = "normal_dark"
        else:
            style = "normal_light"
        write(row, 0, result.sample.name, s=style)
        write(row, 1, result.distribution_type.name, s=style)
        write(row, 2, len(result), s=style)
        write(row, 3, "Default" if result.settings is None else str(result.settings), s=style)
        write(row, 4, "None" if result.x0 is None else str(result.x0.tolist()), s=style)
        write(row, 5, str(result.parameters[-1].tolist()), s=style)
        write(row, 6, result.time_spent, s=style)
        write(row, 7, len(result.loss_series("lmse")), s=style)
        write(row, 8, result.loss("lmse"), s=style)
        if progress_callback is not None:
            progress_callback(i / len(dataset) * 0.1 + 0.1)

    logger.debug("Creating the Statistical Moments sheet.")
    ws = wb.create_sheet("Statistical Moments")
    write(0, 0, "Sample Name", s="header")
    ws.merge_cells(start_row=1, start_column=1, end_row=2, end_column=1)
    ws.column_dimensions[column_to_char(0)].width = 16
    headers = []
    sub_headers = ["Proportion [%]",
                   "Mean [phi]",
                   "Mean [microns]",
                   "Standard Deviation [phi]",
                   "Standard Deviation [microns]",
                   "Skewness",
                   "Kurtosis"]
    for i in range(max_n_components):
        write(0, i * len(sub_headers) + 1, f"C{i+1}", s="header")
        ws.merge_cells(start_row=1, start_column=i*len(sub_headers)+2, end_row=1, end_column=(i+1)*len(sub_headers)+1)
        headers.extend(sub_headers)
    for col, value in enumerate(headers, 1):
        write(1, col, value, s="header")
        ws.column_dimensions[column_to_char(col)].width = 10
    flag_index = 0
    for i, result in enumerate(results):
        row = i + 2
        if row % 2 == 0:
            style = "normal_light"
        else:
            style = "normal_dark"
        write(row, 0, result.sample.name, s=style)
        for component in result:
            index = flags[flag_index]
            s = all_statistics(result.classes, result.classes_phi, component.distribution)
            write(row, index * len(sub_headers) + 1, component.proportion*100, s=style)
            write(row, index * len(sub_headers) + 2, s["logarithmic"]["mean"], s=style)
            write(row, index * len(sub_headers) + 3, s["geometric"]["mean"], s=style)
            write(row, index * len(sub_headers) + 4, s["logarithmic"]["std"], s=style)
            write(row, index * len(sub_headers) + 5, s["geometric"]["std"], s=style)
            write(row, index * len(sub_headers) + 6, s["logarithmic"]["skewness"], s=style)
            write(row, index * len(sub_headers) + 7, s["logarithmic"]["kurtosis"], s=style)
            flag_index += 1
        if progress_callback is not None:
            progress_callback(i / len(dataset) * 0.1 + 0.2)

    logger.debug("Creating the Unmixed Components sheet.")
    ws = wb.create_sheet("Unmixed Components")
    ws.merge_cells(start_row=1, start_column=1, end_row=1, end_column=2)
    write(0, 0, "Sample Name", s="header")
    ws.column_dimensions[column_to_char(0)].width = 16
    for col, value in enumerate(dataset.classes, 2):
        write(0, col, value, s="header")
        ws.column_dimensions[column_to_char(col)].width = 10
    row = 1
    for i, result in enumerate(results):
        if i % 2 == 0:
            style = "normal_light"
        else:
            style = "normal_dark"
        write(row, 0, result.sample.name, s=style)
        ws.merge_cells(start_row=row+1, start_column=1, end_row=row+len(result)+1, end_column=1)
        for component_i, component in enumerate(result, 1):
            write(row, 1, f"C{component_i}", s=style)
            for col, value in enumerate(component.distribution*component.proportion, 2):
                write(row, col, value, s=style)
            row += 1
        write(row, 1, "Sum", s=style)
        for col, value in enumerate(result.distribution, 2):
            write(row, col, value, s=style)
        row += 1
        if progress_callback is not None:
            progress_callback(i / len(dataset) * 0.2 + 0.3)

    logger.debug("Creating separate sheets for all components.")
    ws_dict = {}
    flag_set = set(flags)
    for flag in flag_set:
        ws = wb.create_sheet(f"C{flag+1}")
        write(0, 0, "Sample Name", s="header")
        ws.column_dimensions[column_to_char(0)].width = 16
        for col, value in enumerate(dataset.classes, 1):
            write(0, col, value, s="header")
            ws.column_dimensions[column_to_char(col)].width = 10
        ws_dict[flag] = ws

    flag_index = 0
    for i, result in enumerate(results):
        row = i + 1
        if row % 2 == 0:
            style = "normal_dark"
        else:
            style = "normal_light"

        for component in result:
            flag = flags[flag_index]
            ws = ws_dict[flag]
            write(row, 0, result.sample.name, s=style)
            for col, value in enumerate(component.distribution, 1):
                write(row, col, value, s=style)
            flag_index += 1
        if progress_callback is not None:
            progress_callback(i / len(dataset) * 0.5 + 0.5)

    wb.save(filename)
    wb.close()
    if progress_callback is not None:
        progress_callback(1.0)
    logger.info(f"The SSU results have been saved to the Excel file: [{filename}].")


def save_udm(result: UDMResult, filename: str, progress_callback: Callable[[float], None] = None,
             logger: logging.Logger = None):
    if logger is None:
        logger = logging.getLogger("QGrain")
    else:
        assert isinstance(logger, logging.Logger)
    logger.debug("Start to save the UDM result.")

    wb = openpyxl.Workbook()
    prepare_styles(wb)
    logger.debug("Creating the README sheet.")
    readme_text = \
        f"""
        It contains {result.n_components+4} sheets:
        1. The first sheet is used to put the grain size distributions of corresponding samples.
        2. The second sheet is used to put the resolved parameters of all samples.
        3. The third sheet stores the statistical parameters of all components.
        4. The fourth sheet is used to put the distributions of unmixed components and the sum.
        5-{result.n_components+4}. Other sheets severally store the distributions of each component group.

        The UDM algorithm was proposed and implemented by QGrain.

        UDM Algorithm Details
            Number of Samples: {result.n_samples}
            Kernel Type: {result.kernel_type.name}
            Number of Components: {result.n_components}
            x0: {result.x0 if result.x0 is None else result.x0.tolist()}
            Number of Epochs: {len(result.loss_series("total"))}
            Spent Time: {result.time_spent} s
            
            Computing Device: {"Unknown" if result.settings is None else result.settings["device"]}
            Number of Pretrain Epochs: {"Unknown" if result.settings is None else result.settings["pretrain_epochs"]}
            Minimum Number of Epochs: {"Unknown" if result.settings is None else result.settings["min_epochs"]}
            Maximum Number of Epochs: {"Unknown" if result.settings is None else result.settings["max_epochs"]}
            Learning Rate: {"Unknown" if result.settings is None else result.settings["learning_rate"]}
            Precision: {"Unknown" if result.settings is None else result.settings["precision"]}
            Betas: {"Unknown" if result.settings is None else result.settings["betas"]}
            Constraint Level: {"Unknown" if result.settings is None else result.settings["constraint_level"]}
            Need History: {"Unknown" if result.settings is None else result.settings["need_history"]}
        """

    _write_readme_sheet(wb.active, readme_text)
    logger.debug("Creating the GSDs sheet.")
    if progress_callback is not None:
        _callback = lambda p: progress_callback(p * 0.1)
    else:
        _callback = None
    ws = wb.create_sheet("GSDs")
    _write_dataset_sheet(ws, result.dataset, progress_callback=_callback)

    def write(r, c, v, s="normal_light"):
        cell = ws.cell(r + 1, c + 1, value=v)
        cell.style = s

    logger.debug("Creating the Resolved Parameters sheet.")
    ws = wb.create_sheet("Resolved Parameters")
    write(0, 0, "Sample Name", s="header")
    ws.merge_cells(start_row=1, start_column=1, end_row=2, end_column=1)
    ws.column_dimensions[column_to_char(0)].width = 16
    headers = []
    distribution_type = DistributionType._member_map_[result.kernel_type.name]
    distribution_class = get_distribution(distribution_type)
    sub_headers = (*distribution_class.PARAMETER_NAMES, "Weight")
    for i in range(result.n_components):
        write(0, i * len(sub_headers) + 1, f"C{i+1}", s="header")
        ws.merge_cells(start_row=1, start_column=i*len(sub_headers)+2, end_row=1, end_column=(i+1)*len(sub_headers)+1)
        headers.extend(sub_headers)
    for col, value in enumerate(headers, 1):
        write(1, col, value, s="header")
        ws.column_dimensions[column_to_char(col)].width = 10
    for i, sample in enumerate(result.dataset):
        row = i + 2
        if row % 2 == 0:
            style = "normal_light"
        else:
            style = "normal_dark"
        write(row, 0, sample.name, s=style)
        for j in range(result.n_components):
            for k in range(len(sub_headers)):
                col = j * len(sub_headers) + k + 1
                value = result.parameters[-1, i, k, j]
                write(row, col, value, s=style)
        if progress_callback is not None:
            progress_callback(i / result.n_samples * 0.1 + 0.1)

    logger.debug("Creating the Statistical Moments sheet.")
    ws = wb.create_sheet("Statistical Moments")
    write(0, 0, "Sample Name", s="header")
    ws.merge_cells(start_row=1, start_column=1, end_row=2, end_column=1)
    ws.column_dimensions[column_to_char(0)].width = 16
    headers = []
    sub_headers = ["Proportion [%]",
                   "Mean [phi]",
                   "Mean [microns]",
                   "Standard Deviation [phi]",
                   "Standard Deviation [microns]",
                   "Skewness",
                   "Kurtosis"]
    for i in range(result.n_components):
        write(0, i * len(sub_headers) + 1, f"C{i+1}", s="header")
        ws.merge_cells(start_row=1, start_column=i*len(sub_headers)+2, end_row=1, end_column=(i+1)*len(sub_headers)+1)
        headers.extend(sub_headers)
    for col, value in enumerate(headers, 1):
        write(1, col, value, s="header")
        ws.column_dimensions[column_to_char(col)].width = 10

    for i, sample in enumerate(result.dataset):
        row = i + 2
        if row % 2 == 0:
            style = "normal_light"
        else:
            style = "normal_dark"
        write(row, 0, sample.name, s=style)
        for j in range(result.n_components):
            s = all_statistics(result.dataset.classes, result.dataset.classes_phi, result.components[i, j])
            write(row, j * len(sub_headers) + 1, result.proportions[i, 0, j] * 100, s=style)
            write(row, j * len(sub_headers) + 2, s["logarithmic"]["mean"], s=style)
            write(row, j * len(sub_headers) + 3, s["geometric"]["mean"], s=style)
            write(row, j * len(sub_headers) + 4, s["logarithmic"]["std"], s=style)
            write(row, j * len(sub_headers) + 5, s["geometric"]["std"], s=style)
            write(row, j * len(sub_headers) + 6, s["logarithmic"]["skewness"], s=style)
            write(row, j * len(sub_headers) + 7, s["logarithmic"]["kurtosis"], s=style)
        if progress_callback is not None:
            progress_callback(i / result.n_samples * 0.1 + 0.2)

    logger.debug("Creating the Unmixed Components sheet.")
    ws = wb.create_sheet("Unmixed Components")
    ws.merge_cells(start_row=1, start_column=1, end_row=1, end_column=2)
    write(0, 0, "Sample Name", s="header")
    ws.column_dimensions[column_to_char(0)].width = 16
    for col, value in enumerate(result.dataset.classes, 2):
        write(0, col, value, s="header")
        ws.column_dimensions[column_to_char(col)].width = 10

    predict = result.proportions @ result.components
    row = 1
    for i, sample in enumerate(result.dataset):
        if i % 2 == 0:
            style = "normal_light"
        else:
            style = "normal_dark"
        write(row, 0, sample.name, s=style)
        ws.merge_cells(start_row=row+1, start_column=1, end_row=row+result.n_components+1, end_column=1)
        for j in range(result.n_components):
            write(row, 1, f"C{j+1}", s=style)
            for col, value in enumerate(result.components[i, j]*result.proportions[i, 0, j], 2):
                write(row, col, value, s=style)
            row += 1
        write(row, 1, "Sum", s=style)
        for col, value in enumerate(predict[i, 0], 2):
            write(row, col, value, s=style)
        row += 1
        if progress_callback is not None:
            progress_callback(i / result.n_samples * 0.2 + 0.3)

    logger.debug("Creating separate sheets for all components.")
    for j in range(result.n_components):
        ws = wb.create_sheet(f"C{j+1}")
        write(0, 0, "Sample Name", s="header")
        ws.column_dimensions[column_to_char(0)].width = 16
        for col, value in enumerate(result.dataset.classes, 1):
            write(0, col, value, s="header")
            ws.column_dimensions[column_to_char(col)].width = 10
        for i, sample in enumerate(result.dataset):
            row = i + 1
            if row % 2 == 0:
                style = "normal_dark"
            else:
                style = "normal_light"
            write(row, 0, sample.name, s=style)
            for col, value in enumerate(result.components[i, j], 1):
                write(row, col, value, s=style)
            if progress_callback is not None:
                progress_callback(((i / result.n_samples) + j) / result.n_components * 0.5 + 0.5)

    wb.save(filename)
    wb.close()
    if progress_callback is not None:
        progress_callback(1.0)
    logger.info(f"The UDM result have been saved to the Excel file: [{filename}].")



__all__ = ["SSUResultViewer"]

import logging
import pickle
from typing import *

import numpy as np
from PySide6 import QtCore, QtGui, QtWidgets

from ..models import SSUResult
from ..ssu import built_in_losses
from ..metrics import loss_numpy
from ..charts.BoxplotChart import BoxplotChart
from ..charts.LossSeriesChart import LossSeriesChart
from ..io import save_ssu
from .ParameterTable import ParameterTable


class SSUResultViewer(QtWidgets.QWidget):
    PAGE_ROWS = 20
    logger = logging.getLogger("QGrain.SSUResultViewer")
    result_displayed = QtCore.Signal(SSUResult)
    result_referred = QtCore.Signal(SSUResult)

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self._results: List[SSUResult] = []
        self.setWindowTitle(self.tr("SSU Result Viewer"))
        self.data_table = QtWidgets.QTableWidget(100, 100)
        self.data_table.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.data_table.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.data_table.setAlternatingRowColors(True)
        self.data_table.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.main_layout = QtWidgets.QGridLayout(self)
        self.main_layout.addWidget(self.data_table, 0, 0, 1, 3)
        self.previous_button = QtWidgets.QPushButton(self.tr("Previous"))
        self.previous_button.setToolTip(self.tr("Click to get back to the previous page."))
        self.page_combo_box = QtWidgets.QComboBox()
        self.page_combo_box.addItem(self.tr("Page {0}").format(1))
        self.next_button = QtWidgets.QPushButton(self.tr("Next"))
        self.next_button.setToolTip(self.tr("Click to jump to the next page."))
        self.previous_button.clicked.connect(lambda: self.page_combo_box.setCurrentIndex(max(self.page_index-1, 0)))
        self.page_combo_box.currentIndexChanged.connect(self.update_page)
        self.next_button.clicked.connect(lambda: self.page_combo_box.setCurrentIndex(min(self.page_index+1, self.n_pages-1)))
        self.main_layout.addWidget(self.previous_button, 1, 0)
        self.main_layout.addWidget(self.page_combo_box, 1, 1)
        self.main_layout.addWidget(self.next_button, 1, 2)
        self.loss_label = QtWidgets.QLabel(self.tr("Loss"))
        self.loss_label.setToolTip(self.tr("The function to calculate the difference between prediction and observation."))
        self.loss_combo_box = QtWidgets.QComboBox()
        self.loss_combo_box.addItems(built_in_losses)
        self.loss_combo_box.setCurrentText("lmse")
        self.loss_combo_box.currentTextChanged.connect(lambda: self.update_page(self.page_index))
        self.main_layout.addWidget(self.loss_label, 2, 0)
        self.main_layout.addWidget(self.loss_combo_box, 2, 1, 1, 2)
        self.menu = QtWidgets.QMenu(self.data_table)
        self.menu.setShortcutAutoRepeat(True)
        self.remove_action = self.menu.addAction(self.tr("Remove"))
        self.remove_action.triggered.connect(self.remove_selections)
        self.remove_all_action = self.menu.addAction(self.tr("Remove All"))
        self.remove_all_action.triggered.connect(self.remove_all_results)
        self.refer_action = self.menu.addAction(self.tr("Refer Parameters"))
        self.refer_action.triggered.connect(self.refer_result)
        self.show_chart_action = self.menu.addAction(self.tr("Show Chart"))
        self.show_chart_action.triggered.connect(self.show_chart)
        self.auto_show_selected_action = self.menu.addAction(self.tr("Auto Show"))
        self.auto_show_selected_action.setCheckable(True)
        self.auto_show_selected_action.setChecked(False)
        self.show_distance_action = self.menu.addAction(self.tr("Show Loss Series"))
        self.show_distance_action.triggered.connect(self.show_loss_series)
        self.show_parameter_action = self.menu.addAction(self.tr("Show Parameters"))
        self.show_parameter_action.triggered.connect(self.show_parameters)
        self.detect_outliers_menu = self.menu.addMenu(self.tr("Check"))
        self.check_nan_and_inf_action = self.detect_outliers_menu.addAction(self.tr("NaN / Inf"))
        self.check_nan_and_inf_action.triggered.connect(self.check_nan_and_inf)
        self.check_final_distances_action = self.detect_outliers_menu.addAction(self.tr("Final Loss"))
        self.check_final_distances_action.triggered.connect(self.check_final_distances)
        self.check_mean_action = self.detect_outliers_menu.addAction(self.tr("Mean"))
        self.check_mean_action.triggered.connect(lambda: self.check_component_moments("mean"))
        self.check_std_action = self.detect_outliers_menu.addAction(self.tr("Sorting Coefficient"))
        self.check_std_action.triggered.connect(lambda: self.check_component_moments("std"))
        self.check_skewness_action = self.detect_outliers_menu.addAction(self.tr("Skewness"))
        self.check_skewness_action.triggered.connect(lambda: self.check_component_moments("skewness"))
        self.check_kurtosis_action = self.detect_outliers_menu.addAction(self.tr("Kurtosis"))
        self.check_kurtosis_action.triggered.connect(lambda: self.check_component_moments("kurtosis"))
        self.check_proportion_action = self.detect_outliers_menu.addAction(self.tr("Proportion"))
        self.check_proportion_action.triggered.connect(self.check_component_proportion)
        self.data_table.customContextMenuRequested.connect(self.show_menu)
        self.data_table.itemSelectionChanged.connect(self.on_selection_changed)
        # necessary to add actions of menu to this widget itself,
        # otherwise, the shortcuts will not be triggered
        self.addActions(self.menu.actions())

        self.boxplot_chart = BoxplotChart()
        self.loss_chart = LossSeriesChart()
        self.update_page_list()
        self.update_page(self.page_index)
        self.normal_msg = QtWidgets.QMessageBox(self)
        self.parameter_table = None

    def show_menu(self, pos: QtCore.QPoint):
        self.menu.popup(QtGui.QCursor.pos())

    def show_message(self, title: str, message: str):
        self.normal_msg.setWindowTitle(title)
        self.normal_msg.setText(message)
        self.normal_msg.exec_()

    def show_info(self, message: str):
        self.show_message(self.tr("Info"), message)

    def show_warning(self, message: str):
        self.show_message(self.tr("Warning"), message)

    def show_error(self, message: str):
        self.show_message(self.tr("Error"), message)

    @property
    def loss_name(self) -> str:
        return self.loss_combo_box.currentText()

    @property
    def distance_function(self) -> Callable:
        return loss_numpy(self.loss_combo_box.currentText())

    @property
    def page_index(self) -> int:
        return self.page_combo_box.currentIndex()

    @property
    def n_pages(self) -> int:
        return self.page_combo_box.count()

    @property
    def n_results(self) -> int:
        return len(self._results)

    @property
    def selections(self) -> List[int]:
        start = self.page_index*self.PAGE_ROWS
        temp = set()
        for item in self.data_table.selectedRanges():
            for i in range(item.topRow(), min(self.PAGE_ROWS+1, item.bottomRow()+1)):
                temp.add(i+start)
        indexes = list(temp)
        indexes.sort()
        return indexes

    @property
    def all_results(self) -> List[SSUResult]:
        return self._results.copy()

    @property
    def auto_show_selected(self) -> bool:
        return self.auto_show_selected_action.isChecked()

    def update_page_list(self):
        last_page_index = self.page_index
        if self.n_results == 0:
            n_pages = 1
        else:
            n_pages, left = divmod(self.n_results, self.PAGE_ROWS)
            if left != 0:
                n_pages += 1
        self.page_combo_box.blockSignals(True)
        self.page_combo_box.clear()
        self.page_combo_box.addItems([self.tr("Page {0}").format(i+1) for i in range(n_pages)])
        if last_page_index >= n_pages:
            self.page_combo_box.setCurrentIndex(n_pages-1)
        else:
            self.page_combo_box.setCurrentIndex(last_page_index)
        self.page_combo_box.blockSignals(False)

    def update_page(self, page_index: int):
        def write(row: int, col: int, value: Union[int, float, str]):
            if isinstance(value, str):
                pass
            elif isinstance(value, int):
                value = str(value)
            elif isinstance(value, float):
                value = f"{value: 0.4f}"
            else:
                value = value.__str__()
            item = QtWidgets.QTableWidgetItem(value)
            item.setTextAlignment(QtCore.Qt.AlignCenter)
            self.data_table.setItem(row, col, item)
        # necessary to clear
        self.data_table.clear()
        if page_index == self.n_pages - 1:
            start = page_index * self.PAGE_ROWS
            end = self.n_results
        else:
            start, end = page_index * self.PAGE_ROWS, (page_index+1) * self.PAGE_ROWS
        self.data_table.setRowCount(end-start)
        self.data_table.setColumnCount(6)
        self.data_table.setHorizontalHeaderLabels([
            self.tr("Distribution Type"),
            self.tr("Number of Components"),
            self.tr("Number of Iterations"),
            self.tr("Spent Time [s]"),
            self.tr("Final Loss"),
            self.tr("Has Reference")])
        self.data_table.horizontalHeader().setDefaultAlignment(QtCore.Qt.AlignCenter | QtCore.Qt.TextWordWrap)
        sample_names = [result.sample.name for result in self._results[start: end]]
        self.data_table.setVerticalHeaderLabels(sample_names)
        for row, result in enumerate(self._results[start: end]):
            write(row, 0, result.distribution_type.name)
            write(row, 1, len(result))
            write(row, 2, result.n_iterations)
            write(row, 3, result.time_spent)
            write(row, 4, result.loss(self.loss_name))
            write(row, 5, self.tr("Yes") if result.x0 is not None else self.tr("No"))
        self.data_table.resizeColumnsToContents()

    def add_result(self, result: SSUResult):
        if self.n_results == 0 or (self.page_index==self.n_pages-1 and divmod(self.n_results, self.PAGE_ROWS)[-1]!=0):
            need_update = True
        else:
            need_update = False
        self._results.append(result)
        self.update_page_list()
        if need_update:
            self.update_page(self.page_index)

    def add_results(self, results: List[SSUResult]):
        if self.n_results == 0 or (self.page_index==self.n_pages-1 and divmod(self.n_results, self.PAGE_ROWS)[-1]!=0):
            need_update = True
        else:
            need_update = False
        self._results.extend(results)
        self.update_page_list()
        if need_update:
            self.update_page(self.page_index)

    def remove_results(self, indexes):
        results = []
        for i in reversed(indexes):
            res = self._results.pop(i)
            results.append(res)
        self.update_page_list()
        self.update_page(self.page_index)

    def remove_selections(self):
        indexes = self.selections
        self.remove_results(indexes)

    def remove_all_results(self):
        remove_msg = QtWidgets.QMessageBox(self)
        remove_msg.setStandardButtons(QtWidgets.QMessageBox.No | QtWidgets.QMessageBox.Yes)
        remove_msg.setDefaultButton(QtWidgets.QMessageBox.No)
        remove_msg.setWindowTitle(self.tr("Warning"))
        remove_msg.setText(self.tr("Are you sure to remove all SSU results?"))
        res = remove_msg.exec_()
        if res == QtWidgets.QMessageBox.Yes:
            self._results.clear()
            self.update_page_list()
            self.update_page(0)

    def refer_result(self):
        results = [self._results[i] for i in self.selections]
        if results is None or len(results) == 0:
            return
        result = results[0]
        self.result_referred.emit(result)

    def show_chart(self):
        results = [self._results[i] for i in self.selections]
        if results is None or len(results) == 0:
            return
        result = results[0]
        self.result_displayed.emit(result)

    def on_selection_changed(self):
        if self.auto_show_selected:
            self.show_chart()

    def show_loss_series(self):
        results = [self._results[i] for i in self.selections]
        if results is None or len(results) == 0:
            return
        result = results[0]
        series = result.loss_series(self.loss_name)
        if len(series) == 1:
            self.show_warning(self.tr("The variation history was not recorded, can not calculate the loss series."))
            return
        self.loss_chart.show_loss_series(series, self.loss_name, result.sample.name)
        self.loss_chart.show()

    def show_parameters(self):
        results = [self._results[i] for i in self.selections]
        if results is None or len(results) == 0:
            return
        result = results[0]
        self.parameter_table = ParameterTable(result)
        self.parameter_table.show()

    def load_results(self):
        filename, _ = QtWidgets.QFileDialog.getOpenFileName(
            self, self.tr("Choose the file which stores the dumped SSU results"),
            ".", "Dumped SSU Results (*.ssu)")
        if filename is None or filename == "":
            return
        with open(filename, "rb") as f:
            results = pickle.load(f) # type: list[SSUResult]
            valid = True
            if isinstance(results, list):
                for result in results:
                    if not isinstance(result, SSUResult):
                        valid = False
                        break
            else:
                valid = False

            if valid:
                if self.n_results != 0 and len(results) != 0:
                    old_classes = self._results[0].classes_phi
                    new_classes = results[0].classes_phi
                    classes_inconsistent = False
                    if len(old_classes) != len(new_classes):
                        classes_inconsistent = True
                    else:
                        classes_error = np.abs(old_classes - new_classes)
                        if not np.all(np.less_equal(classes_error, 1e-8)):
                            classes_inconsistent = True
                    if classes_inconsistent:
                        self.logger.error("The grain size classes of the SSU results in binary file are inconsistent with that in list.")
                        self.show_error(self.tr("The grain size classes of the SSU results in binary file are inconsistent with that in list."))
                        return
                self.add_results(results)
                self.logger.info(f"There are {len(results)} SSU results that have been loaded.")
            else:
                self.logger.error("The binary file is invalid (i.e., the objects in it are not SSU results).")
                self.show_error(self.tr("The binary file is invalid (i.e., the objects in it are not SSU results)."))

    def save_results(self, align_components=False):
        if self.n_results == 0:
            self.show_error(self.tr("There is no SSU result."))
            return
        filename, _ = QtWidgets.QFileDialog.getSaveFileName(
            self, self.tr("Choose a filename to save the SSU Results"),
            ".", "Microsoft Excel (*.xlsx);;Dumped SSU Results (*.ssu)")
        if filename is None or filename == "":
            return
            # Excel
        if filename[-4:] == "xlsx":
            progress_dialog = QtWidgets.QProgressDialog(
                    self.tr("Saving the SSU results..."), self.tr("Cancel"),
                    0, 100, self)
            progress_dialog.setWindowTitle("QGrain")
            progress_dialog.setWindowModality(QtCore.Qt.WindowModal)

            def callback(progress: float):
                if progress_dialog.wasCanceled():
                    raise StopIteration()
                progress_dialog.setValue(int(progress*100))
                QtCore.QCoreApplication.processEvents()
            try:
                save_ssu(self._results, filename, align_components, progress_callback=callback, logger=self.logger)
            except StopIteration as e:
                self.logger.info("The saving task was canceled.")
            finally:
                progress_dialog.close()
        else:
            with open(filename, "wb") as f:
                pickle.dump(self._results, f)
                self.logger.info("All SSU results have been dumped.")

    def ask_deal_outliers(self, outlier_results: List[SSUResult], outlier_indexes: List[int]):
        assert len(outlier_indexes) == len(outlier_results)
        if len(outlier_results) == 0:
            self.logger.info("There is no result was evaluated as the outlier.")
        else:
            self.logger.info(f"There are {len(outlier_results)} results were evaluated as the outliers: {', '.join([result.sample.name for result in outlier_results])}.")
            outlier_msg = QtWidgets.QMessageBox(self)
            outlier_msg.setStandardButtons(QtWidgets.QMessageBox.No | QtWidgets.QMessageBox.Yes)
            outlier_msg.setDefaultButton(QtWidgets.QMessageBox.No)
            outlier_msg.setText(self.tr("There are {0} results were evaluated as the outliers. Please check the logs for more details. Do you want to remove them?").format(len(outlier_results)))
            res = outlier_msg.exec_()
            if res == QtWidgets.QMessageBox.Yes:
                self.remove_results(outlier_indexes)
                self.logger.debug("The outliers have been removed.")

    def check_nan_and_inf(self):
        if self.n_results == 0:
            self.show_error(self.tr("There is no SSU result."))
            return
        outlier_results = []
        outlier_indexes = []
        for i, result in enumerate(self._results):
            if not result.is_valid:
                outlier_results.append(result)
                outlier_indexes.append(i)
        self.logger.debug("Check if there is any NaN or Inf value.")
        self.ask_deal_outliers(outlier_results, outlier_indexes)

    def check_final_distances(self):
        if self.n_results == 0:
            self.show_error(self.tr("There is no SSU result."))
            return
        elif self.n_results < 10:
            self.show_error(self.tr("The number of results is not enough."))
            return
        losses = []
        for result in self._results:
            losses.append(result.loss(self.loss_name))
        losses = np.array(losses)
        self.boxplot_chart.show_dataset([losses], xlabels=[self.loss_name], ylabel="Loss")
        self.boxplot_chart.show()

        # calculate the 1/4, 1/2, and 3/4 position value to judge which result is invalid
        # 1. the mean squared errors are much higher in the results which are lack of components
        # 2. with the component number getting higher, the mean squared error will get lower and finally reach the minimum
        median = np.median(losses)
        upper_group = losses[np.greater(losses, median)]
        lower_group = losses[np.less(losses, median)]
        value_1_4 = np.median(lower_group)
        value_3_4 = np.median(upper_group)
        loss_QR = value_3_4 - value_1_4
        outlier_results = []
        outlier_indexes = []
        for i, (result, loss) in enumerate(zip(self._results, losses)):
            if loss > value_3_4 + loss_QR * 1.5:
            # which error too small is not outlier
            # if loss > value_3_4 + loss_QR * 1.5 or loss < value_1_4 - loss_QR * 1.5:
                outlier_results.append(result)
                outlier_indexes.append(i)
        self.logger.debug(f"Check the final losses using Whisker plot.")
        self.ask_deal_outliers(outlier_results, outlier_indexes)

    def check_component_moments(self, key: str):
        if self.n_results == 0:
            self.show_error(self.tr("There is no SSU result."))
            return
        elif self.n_results < 10:
            self.show_error(self.tr("The number of results is not enough."))
            return
        max_n_components = 0
        for result in self._results:
            if len(result) > max_n_components:
                max_n_components = len(result)
        moments = []
        for i in range(max_n_components):
            moments.append([])

        for result in self._results:
            for i, component in enumerate(result):
                if np.isnan(component.moments[key]) or np.isinf(component.moments[key]):
                    pass
                else:
                    moments[i].append(component.moments[key])

        key_label_trans = {"mean": "Mean [φ]", "std": "Sorting Coefficient", "skewness": "Skewness", "kurtosis": "Kurtosis"}
        self.boxplot_chart.show_dataset(moments, xlabels=[f"C{i+1}" for i in range(max_n_components)], ylabel=key_label_trans[key])
        self.boxplot_chart.show()

        outlier_dict = {}

        for i in range(max_n_components):
            stacked_moments = np.array(moments[i])
            # calculate the 1/4, 1/2, and 3/4 position value to judge which result is invalid
            # 1. the mean squared errors are much higher in the results which are lack of components
            # 2. with the component number getting higher, the mean squared error will get lower and finally reach the minimum
            median = np.median(stacked_moments)
            upper_group = stacked_moments[np.greater(stacked_moments, median)]
            lower_group = stacked_moments[np.less(stacked_moments, median)]
            value_1_4 = np.median(lower_group)
            value_3_4 = np.median(upper_group)
            moment_QR = value_3_4 - value_1_4

            for j, result in enumerate(self._results):
                if len(result) > i:
                    moment = result[i].moments[key]
                    if moment > value_3_4 + moment_QR * 1.5 or moment < value_1_4 - moment_QR * 1.5:
                        outlier_dict[j] = result

        outlier_results = []
        outlier_indexes = []
        for index, result in sorted(outlier_dict.items(), key=lambda x:x[0]):
            outlier_indexes.append(index)
            outlier_results.append(result)
        self.logger.debug(f"Check the {key_label_trans[key]} values using Whisker plot.")
        self.ask_deal_outliers(outlier_results, outlier_indexes)

    def check_component_proportion(self):
        outlier_results = []
        outlier_indexes = []
        for i, result in enumerate(self._results):
            for component in result:
                if component.proportion < 1e-3:
                    outlier_results.append(result)
                    outlier_indexes.append(i)
                    break
        self.logger.debug("Check if the proportion of any component is near zero.")
        self.ask_deal_outliers(outlier_results, outlier_indexes)

    def changeEvent(self, event: QtCore.QEvent):
        if event.type() == QtCore.QEvent.LanguageChange:
            self.retranslate()

    def keyPressEvent(self, event: QtGui.QKeyEvent):
        if event.key() == QtCore.Qt.Key_Up:
            self.page_combo_box.setCurrentIndex(max(self.page_index-1, 0))
        elif event.key() == QtCore.Qt.Key_Down:
            self.page_combo_box.setCurrentIndex(min(self.page_index+1, self.n_pages-1))

    def retranslate(self):
        self.setWindowTitle(self.tr("SSU Result Viewer"))
        self.previous_button.setText(self.tr("Previous"))
        self.previous_button.setToolTip(self.tr("Click to get back to the previous page."))
        for i in range(self.page_combo_box.count()):
            self.page_combo_box.setItemText(i, self.tr("Page {0}").format(i+1))
        self.next_button.setText(self.tr("Next"))
        self.next_button.setToolTip(self.tr("Click to jump to the next page."))
        self.loss_label.setText(self.tr("Loss"))
        self.loss_label.setToolTip(self.tr("The function to calculate the difference between prediction and observation."))
        self.remove_action.setText(self.tr("Remove"))
        self.remove_all_action.setText(self.tr("Remove All"))
        self.refer_action.setText(self.tr("Refer Parameters"))
        self.show_chart_action.setText(self.tr("Show Chart"))
        self.auto_show_selected_action.setText(self.tr("Auto Show"))
        self.show_distance_action.setText(self.tr("Show Loss Series"))
        self.show_parameter_action.setText(self.tr("Show Parameters"))
        self.detect_outliers_menu.setTitle(self.tr("Check"))
        self.check_nan_and_inf_action.setText(self.tr("NaN / Inf"))
        self.check_final_distances_action.setText(self.tr("Final Loss"))
        self.check_mean_action.setText(self.tr("Mean"))
        self.check_std_action.setText(self.tr("Sorting Coefficient"))
        self.check_skewness_action.setText(self.tr("Skewness"))
        self.check_kurtosis_action.setText(self.tr("Kurtosis"))
        self.check_proportion_action.setText(self.tr("Proportion"))
        self.update_page(self.page_index)



__all__ = ["RuntimeLog", "StatusBarLogHandler", "GUILogHandler"]

import logging
from queue import Queue

from PySide6 import QtCore, QtWidgets

SUCCESS_COLOR = "#55bb8a"
WARNING_COLOR = "#fbda41"
ERROR_COLOR = "#c04851"


class RuntimeLog(QtWidgets.QDialog):
    MAX_SIZE = 200

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.setWindowTitle(self.tr("Runtime Log"))
        self.setMinimumSize(400, 400)
        self.main_layout = QtWidgets.QGridLayout(self)
        self.text = QtWidgets.QPlainTextEdit()
        self.text.setReadOnly(True)
        self.text.setMaximumBlockCount(self.MAX_SIZE)
        self.main_layout.addWidget(self.text, 0, 0)
        self.record_queue = Queue(maxsize=self.MAX_SIZE)

    @classmethod
    def get_html_message(cls, record: logging.LogRecord, message: str):
        format_str = "<font color='{0}'>{1}<font/>"
        if record.levelno < logging.INFO:
            return message
        elif record.levelno < logging.WARNING:
            return format_str.format(SUCCESS_COLOR, message)
        elif record.levelno < logging.ERROR:
            return format_str.format(WARNING_COLOR, message)
        else:
            return format_str.format(ERROR_COLOR, message)

    def add_record(self, record: logging.LogRecord, message: str):
        if self.record_queue.full():
            self.record_queue.get(False)
        self.record_queue.put((record, message), False)
        html_message = self.get_html_message(record, message)
        self.text.appendHtml(html_message)

    def changeEvent(self, event: QtCore.QEvent):
        if event.type() == QtCore.QEvent.LanguageChange:
            self.retranslate()

    def retranslate(self):
        self.setWindowTitle(self.tr("Runtime Log"))


class StatusBarLogHandler(logging.Handler):
    def __init__(self, status_bar: QtWidgets.QStatusBar, level=logging.WARNING):
        super().__init__(level=level)
        self.status_bar = status_bar
        self.__mutex = QtCore.QMutex()

    def set_color(self, level: int):
        if level < logging.INFO:
            self.status_bar.setStyleSheet("")
        elif level < logging.WARNING:
            self.status_bar.setStyleSheet("QStatusBar {color: #55bb8a}")
        elif level < logging.ERROR:
            self.status_bar.setStyleSheet("QStatusBar {color: #fbda41}")
        else:
            self.status_bar.setStyleSheet("QStatusBar {color: #c04851}")

    def emit(self, record: logging.LogRecord):
        self.__mutex.lock()
        if record.levelno < self.level:
            return
        self.set_color(record.levelno)
        message = self.format(record)
        self.status_bar.showMessage(message)
        self.__mutex.unlock()


class GUILogHandler(logging.Handler):
    def __init__(self, log_panel: RuntimeLog, level=logging.WARNING):
        super().__init__(level=level)
        self.log_panel = log_panel
        self.__mutex = QtCore.QMutex()

    def emit(self, record: logging.LogRecord):
        self.__mutex.lock()
        if record.levelno < self.level:
            return
        message = self.format(record)
        self.log_panel.add_record(record, message)
        self.__mutex.unlock()



import logging
import os
import string
from typing import *

from PySide6 import QtCore, QtGui, QtWidgets
from qt_material import apply_stylesheet, list_themes

from .. import QGRAIN_ROOT_PATH
from ..models import Dataset
from ..protos.client import QGrainClient
from ..io import save_pca, save_statistics
from ..utils import udm_to_ssu
from . import EXTRA
from .About import About
from .ClusteringAnalyzer import ClusteringAnalyzer
from .DatasetGenerator import DatasetGenerator
from .DatasetLoader import DatasetLoader
from .EMMAAnalyzer import EMMAAnalyzer
from .EMMASettings import EMMASettings
from .RuntimeLog import *
from .PCAAnalyzer import PCAAnalyzer
from .ParameterEditor import ParameterEditor
from .SSUAnalyzer import SSUAnalyzer
from .SSUMulticoreAnalyzer import SSUMulticoreAnalyzer
from .SSUSettings import SSUSettings
from .StatisticalAnalyzer import StatisticalAnalyzer
from .UDMAnalyzer import UDMAnalyzer
from .UDMSettings import UDMSettings


class MainWindow(QtWidgets.QMainWindow):
    logger = logging.getLogger("QGrain.MainWindow")

    def __init__(self):
        super().__init__()
        self._dataset: Optional[Dataset] = None
        self._translator: Optional[QtCore.QTranslator] = None
        self._client = QGrainClient()
        self.setWindowTitle("QGrain")
        self.ssu_setting_dialog = SSUSettings(self)
        self.emma_setting_dialog = EMMASettings(parent=self, client=self._client)
        self.udm_setting_dialog = UDMSettings(parent=self, client=self._client)
        self.parameter_editor = ParameterEditor(self)
        self.ssu_multicore_analyzer = SSUMulticoreAnalyzer(self)
        self.tab_widget = QtWidgets.QTabWidget(self)
        self.tab_widget.setTabPosition(QtWidgets.QTabWidget.West)
        self.setCentralWidget(self.tab_widget)
        self.dataset_generator = DatasetGenerator(self)
        self.tab_widget.addTab(self.dataset_generator, self.tr("Generator"))
        self.dataset_viewer = StatisticalAnalyzer(self)
        self.tab_widget.addTab(self.dataset_viewer, self.tr("Statistics"))
        self.pca_analyzer = PCAAnalyzer(self)
        self.tab_widget.addTab(self.pca_analyzer, self.tr("PCA"))
        self.clustering_analyzer = ClusteringAnalyzer(self)
        self.tab_widget.addTab(self.clustering_analyzer, self.tr("Clustering"))
        self.ssu_analyzer = SSUAnalyzer(self.ssu_setting_dialog, self.parameter_editor, parent=self)
        self.tab_widget.addTab(self.ssu_analyzer, self.tr("SSU"))
        self.emma_analyzer = EMMAAnalyzer(self.emma_setting_dialog, self.parameter_editor, client=self._client, parent=self)
        self.tab_widget.addTab(self.emma_analyzer, self.tr("EMMA"))
        self.udm_analyzer = UDMAnalyzer(self.udm_setting_dialog, self.parameter_editor, client=self._client, parent=self)
        self.tab_widget.addTab(self.udm_analyzer, self.tr("UDM"))

        # Open
        self.open_menu = self.menuBar().addMenu(self.tr("Open"))
        self.open_dataset_action = self.open_menu.addAction(self.tr("Grain Size Dataset"))
        self.open_dataset_action.triggered.connect(lambda: self.load_dataset_dialog.show())
        self.load_ssu_result_action = self.open_menu.addAction(self.tr("SSU Results"))
        self.load_ssu_result_action.triggered.connect(self.ssu_analyzer.result_view.load_results)
        self.load_emma_result_action = self.open_menu.addAction(self.tr("EMMA Result"))
        self.load_emma_result_action.triggered.connect(self.emma_analyzer.load_result)
        self.load_udm_result_action = self.open_menu.addAction(self.tr("UDM Result"))
        self.load_udm_result_action.triggered.connect(self.udm_analyzer.load_result)

        # Save
        self.save_menu = self.menuBar().addMenu(self.tr("Save"))
        self.save_artificial_action = self.save_menu.addAction(self.tr("Artificial Dataset"))
        self.save_artificial_action.triggered.connect(self.dataset_generator.on_save_clicked)
        self.save_statistics_action = self.save_menu.addAction(self.tr("Statistical Result"))
        self.save_statistics_action.triggered.connect(self.on_save_statistics_clicked)
        self.save_pca_action = self.save_menu.addAction(self.tr("PCA Result"))
        self.save_pca_action.triggered.connect(self.on_save_pca_clicked)
        self.save_clustering_action = self.save_menu.addAction(self.tr("Clustering Result"))
        self.save_clustering_action.triggered.connect(self.clustering_analyzer.save_result)
        self.save_ssu_result_action = self.save_menu.addAction(self.tr("SSU Results"))
        self.save_ssu_result_action.triggered.connect(self.ssu_analyzer.result_view.save_results)
        self.save_emma_result_action = self.save_menu.addAction(self.tr("EMMA Result"))
        self.save_emma_result_action.triggered.connect(self.emma_analyzer.save_selected_result)
        self.save_udm_result_action = self.save_menu.addAction(self.tr("UDM Result"))
        self.save_udm_result_action.triggered.connect(self.udm_analyzer.save_selected_result)

        # Config
        self.config_menu = self.menuBar().addMenu(self.tr("Configure"))
        self.config_ssu_action = self.config_menu.addAction(self.tr("SSU Algorithm"))
        self.config_ssu_action.triggered.connect(self.ssu_setting_dialog.show)
        self.config_emma_action = self.config_menu.addAction(self.tr("EMMA Algorithm"))
        self.config_emma_action.triggered.connect(self.emma_setting_dialog.show)
        self.config_udm_action = self.config_menu.addAction(self.tr("UDM Algorithm"))
        self.config_udm_action.triggered.connect(self.udm_setting_dialog.show)

        # Experimental
        self.experimental_menu = self.menuBar().addMenu(self.tr("Experimental"))
        self.ssu_fit_all_action = self.experimental_menu.addAction(
            self.tr("Perform SSU For All Samples"))
        self.ssu_fit_all_action.triggered.connect(self.ssu_fit_all_samples)
        self.convert_udm_to_ssu_action = self.experimental_menu.addAction(
            self.tr("Convert Selected UDM Result To SSU Results"))
        self.convert_udm_to_ssu_action.triggered.connect(self.convert_udm_to_ssu)
        self.save_all_ssu_figures_action = self.experimental_menu.addAction(
            self.tr("Save Figures For All SSU Results"))
        self.save_all_ssu_figures_action.triggered.connect(self.save_all_ssu_figure)

        # Language
        self.language_menu = self.menuBar().addMenu(self.tr("Language"))
        self.language_group = QtGui.QActionGroup(self.language_menu)
        self.language_group.setExclusive(True)
        self.language_actions: List[QtGui.QAction] = []
        for key, name in self.supported_languages:
            action = self.language_group.addAction(name)
            action.setCheckable(True)
            action.triggered.connect(lambda checked=False, language=key: self.switch_language(language))
            self.language_menu.addAction(action)
            self.language_actions.append(action)

        # Theme
        self.theme_menu = self.menuBar().addMenu(self.tr("Theme"))
        self.theme_group = QtGui.QActionGroup(self.theme_menu)
        self.theme_group.setExclusive(True)
        self.default_theme_action = self.theme_group.addAction(self.tr("Default"))
        self.default_theme_action.setCheckable(True)
        self.default_theme_action.setChecked(True)
        app = QtWidgets.QApplication.instance()
        self.default_theme_action.triggered.connect(lambda: apply_stylesheet(app, theme=os.path.join(
            QGRAIN_ROOT_PATH, "assets", "default_theme.xml"), invert_secondary=True, extra=EXTRA))
        self.theme_menu.addAction(self.default_theme_action)
        self.light_theme_action = self.theme_group.addAction(self.tr("Light"))
        self.light_theme_action.setCheckable(True)
        self.light_theme_action.triggered.connect(lambda: apply_stylesheet(app, theme=os.path.join(
            QGRAIN_ROOT_PATH, "assets", "light_theme.xml"), invert_secondary=True, extra=EXTRA))
        self.theme_menu.addAction(self.light_theme_action)
        self.dark_theme_action = self.theme_group.addAction(self.tr("Dark"))
        self.dark_theme_action.setCheckable(True)
        self.dark_theme_action.triggered.connect(lambda: apply_stylesheet(app, theme=os.path.join(
            QGRAIN_ROOT_PATH, "assets", "dark_theme.xml"), invert_secondary=False, extra=EXTRA))
        self.theme_menu.addAction(self.dark_theme_action)
        self.theme_actions: List[QtGui.QAction] = [self.default_theme_action,
                                                   self.light_theme_action,
                                                   self.dark_theme_action]
        # Log
        self.log_action = QtGui.QAction(self.tr("Log"))
        self.log_action.triggered.connect(lambda: self.log_dialog.show())
        self.menuBar().addAction(self.log_action)

        # About
        self.about_action = QtGui.QAction(self.tr("About"))
        self.about_action.triggered.connect(lambda: self.about_dialog.show())
        self.menuBar().addAction(self.about_action)

        # Connect signals
        self.ssu_multicore_analyzer.result_finished.connect(self.ssu_analyzer.result_view.add_result)
        self.load_dataset_dialog = DatasetLoader(self)
        self.load_dataset_dialog.dataset_loaded.connect(self.on_dataset_loaded)
        self.load_dataset_dialog.dataset_loaded.connect(self.dataset_viewer.on_dataset_loaded)
        self.load_dataset_dialog.dataset_loaded.connect(self.pca_analyzer.on_dataset_loaded)
        self.load_dataset_dialog.dataset_loaded.connect(self.clustering_analyzer.on_dataset_loaded)
        self.load_dataset_dialog.dataset_loaded.connect(self.ssu_analyzer.on_dataset_loaded)
        self.load_dataset_dialog.dataset_loaded.connect(self.emma_analyzer.on_dataset_loaded)
        self.load_dataset_dialog.dataset_loaded.connect(self.udm_analyzer.on_dataset_loaded)
        self.log_dialog = RuntimeLog(self)
        self.about_dialog = About(self)
        self.file_dialog = QtWidgets.QFileDialog(parent=self)
        self.normal_msg = QtWidgets.QMessageBox(self)
        self.close_msg = QtWidgets.QMessageBox(self)
        self.close_msg.setWindowTitle(self.tr("Warning"))
        self.close_msg.setText(self.tr("Closing this window will terminate all running tasks, are you sure to close it?"))
        self.close_msg.setStandardButtons(QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
        self.close_msg.setDefaultButton(QtWidgets.QMessageBox.No)

        settings = QtCore.QSettings(os.path.join(os.path.expanduser("~"), "QGrain", "qgrain.ini"),
                                    QtCore.QSettings.Format.IniFormat)
        language = settings.value("language", "en", type=str)
        for (key, name), action in zip(self.supported_languages, self.language_actions):
            if key == language:
                action.setChecked(True)
                break
        theme = settings.value("theme", "default", type=str)
        for (key, name), action in zip(self.supported_themes, self.theme_actions):
            if key == theme:
                action.setChecked(True)
                break

    @property
    def supported_languages(self) -> List[Tuple[str, str]]:
        languages = [("en", "English"),
                     ("zh_CN", "简体中文")]
        return languages

    @property
    def supported_themes(self) -> List[Tuple[str, str]]:
        languages = [("default", self.tr("Default")),
                     ("light", self.tr("Light")),
                     ("dark", self.tr("Dark"))]
        return languages

    @property
    def language(self) -> str:
        for i, language_action in enumerate(self.language_actions):
            if language_action.isChecked():
                key, name = self.supported_languages[i]
                return key

    @property
    def theme(self) -> str:
        for i, theme_action in enumerate(self.theme_actions):
            if theme_action.isChecked():
                theme, name = self.supported_themes[i]
                return theme

    def show_message(self, title: str, message: str):
        self.normal_msg.setWindowTitle(title)
        self.normal_msg.setText(message)
        self.normal_msg.exec_()

    def show_info(self, message: str):
        self.show_message(self.tr("Info"), message)

    def show_warning(self, message: str):
        self.show_message(self.tr("Warning"), message)

    def show_error(self, message: str):
        self.show_message(self.tr("Error"), message)

    def on_dataset_loaded(self, dataset: Dataset):
        if dataset is None:
            return
        self._dataset = dataset

    def ssu_fit_all_samples(self):
        tasks = self.ssu_analyzer.get_all_tasks()
        self.ssu_multicore_analyzer.setup_tasks(tasks)
        self.ssu_multicore_analyzer.exec()

    def convert_udm_to_ssu(self):
        if self.udm_analyzer.n_results == 0:
            self.show_error(self.tr("There is no UDM result."))
            return
        udm_result = self.udm_analyzer.selected_result
        assert udm_result is not None
        progress_dialog = QtWidgets.QProgressDialog(
            self.tr("Converting the selected UDM result to SSU results..."), self.tr("Cancel"),
            0, 100, self)
        progress_dialog.setWindowTitle("QGrain")
        progress_dialog.setWindowModality(QtCore.Qt.WindowModal)

        def callback(progress: float):
            if progress_dialog.wasCanceled():
                raise StopIteration()
            progress_dialog.setValue(int(progress*100))
            QtCore.QCoreApplication.processEvents()
        ssu_results = udm_to_ssu(udm_result, logger=self.logger, progress_callback=callback)
        self.ssu_analyzer.result_view.add_results(ssu_results)

    def save_all_ssu_figure(self):
        if self.ssu_analyzer.result_view.n_results == 0:
            self.logger.error("There is no SSU result.")
            self.show_error(self.tr("There is no SSU result."))
            return
        directory = self.file_dialog.getExistingDirectory(
            self, self.tr("Choose a directory to save the figures for all SSU results"),
            ".", QtWidgets.QFileDialog.ShowDirsOnly)
        if directory is None or directory == "":
            return
        progress_dialog = QtWidgets.QProgressDialog(
            self.tr("Saving the figures for all SSU results..."), self.tr("Cancel"),
            0, 100, self)
        progress_dialog.setWindowTitle("QGrain")
        progress_dialog.setWindowModality(QtCore.Qt.WindowModal)

        def callback(progress: float):
            if progress_dialog.wasCanceled():
                raise StopIteration()
            progress_dialog.setValue(int(progress*100))
            QtCore.QCoreApplication.processEvents()
        try:
            all_results = self.ssu_analyzer.result_view.all_results
            for i, result in enumerate(all_results):
                self.ssu_analyzer.result_chart.show_chart(result)
                image = self.ssu_analyzer.result_chart.grab()
                filename = os.path.join(directory, f"{i}.png")
                image.save(filename)
                callback(i/len(all_results))
            callback(1.0)
        except StopIteration as e:
            self.logger.info("The saving task was canceled.")
        finally:
            progress_dialog.close()

    def on_save_statistics_clicked(self):
        if self._dataset is None:
            self.logger.error("The dataset has not been loaded.")
            self.show_error(self.tr("The dataset has not been loaded."))
            return
        filename, _ = self.file_dialog.getSaveFileName(
            self, self.tr("Choose a filename to save the statistical result"),
            ".", "Microsoft Excel (*.xlsx)")
        if filename is None or filename == "":
            return
        progress_dialog = QtWidgets.QProgressDialog(
            self.tr("Saving the statistical result..."), self.tr("Cancel"),
            0, 100, self)
        progress_dialog.setWindowTitle("QGrain")
        progress_dialog.setWindowModality(QtCore.Qt.WindowModal)

        def callback(progress: float):
            if progress_dialog.wasCanceled():
                raise StopIteration()
            progress_dialog.setValue(int(progress*100))
            QtCore.QCoreApplication.processEvents()
        try:
            save_statistics(self._dataset, filename, progress_callback=callback, logger=self.logger)
        except StopIteration as e:
            self.logger.info("The saving task was canceled.")
        finally:
            progress_dialog.close()

    def on_save_pca_clicked(self):
        if self._dataset is None:
            self.show_error(self.tr("The dataset has not been loaded."))
            return
        filename, _ = self.file_dialog.getSaveFileName(
            self, self.tr("Choose a filename to save the PCA result"),
            ".", "Microsoft Excel (*.xlsx)")
        if filename is None or filename == "":
            return
        progress_dialog = QtWidgets.QProgressDialog(
            self.tr("Saving the PCA result..."), self.tr("Cancel"),
            0, 100, self)
        progress_dialog.setWindowTitle("QGrain")
        progress_dialog.setWindowModality(QtCore.Qt.WindowModal)

        def callback(progress: float):
            if progress_dialog.wasCanceled():
                raise StopIteration()
            progress_dialog.setValue(int(progress*100))
            QtCore.QCoreApplication.processEvents()
        try:
            save_pca(self._dataset, filename, progress_callback=callback, logger=self.logger)
        except StopIteration as e:
            self.logger.info("The saving task was canceled.")
        finally:
            progress_dialog.close()

    def switch_language(self, language: str):
        app = QtWidgets.QApplication.instance()
        if self._translator is not None:
            app.removeTranslator(self._translator)
        translator = QtCore.QTranslator(app)
        translator.load(os.path.join(QGRAIN_ROOT_PATH, "assets", language))
        app.installTranslator(translator)
        self._translator = translator

    def changeEvent(self, event: QtCore.QEvent):
        if event.type() == QtCore.QEvent.LanguageChange:
            self.retranslate()

    def closeEvent(self, event: QtGui.QCloseEvent):
        res = self.close_msg.exec_()
        if res == QtWidgets.QMessageBox.Yes:
            settings = QtCore.QSettings(os.path.join(os.path.expanduser("~"), "QGrain", "qgrain.ini"),
                                        QtCore.QSettings.Format.IniFormat)
            settings.setValue("language", self.language)
            settings.setValue("theme", self.theme)
            self.clustering_analyzer.closeEvent(event)
            event.accept()
        else:
            event.ignore()

    def retranslate(self):
        self.setWindowTitle("QGrain")
        self.close_msg.setWindowTitle(self.tr("Warning"))
        self.close_msg.setText(self.tr("Closing this window will terminate all running tasks, are you sure to close it?"))
        self.open_menu.setTitle((self.tr("Open")))
        self.save_menu.setTitle(self.tr("Save"))
        self.open_dataset_action.setText(self.tr("Grain Size Dataset"))
        self.load_ssu_result_action.setText(self.tr("SSU Results"))
        self.load_emma_result_action.setText(self.tr("EMMA Result"))
        self.load_udm_result_action.setText(self.tr("UDM Result"))
        self.save_artificial_action.setText(self.tr("Artificial Dataset"))
        self.save_statistics_action.setText(self.tr("Statistical Result"))
        self.save_pca_action.setText(self.tr("PCA Result"))
        self.save_clustering_action.setText(self.tr("Clustering Result"))
        self.save_ssu_result_action.setText(self.tr("SSU Results"))
        self.save_emma_result_action.setText(self.tr("EMMA Result"))
        self.save_udm_result_action.setText(self.tr("UDM Result"))
        self.config_menu.setTitle(self.tr("Configure"))
        self.config_ssu_action.setText(self.tr("SSU Algorithm"))
        self.config_emma_action.setText(self.tr("EMMA Algorithm"))
        self.config_udm_action.setText(self.tr("UDM Algorithm"))
        self.experimental_menu.setTitle(self.tr("Experimental"))
        self.ssu_fit_all_action.setText(self.tr("Perform SSU For All Samples"))
        self.convert_udm_to_ssu_action.setText(self.tr("Convert Selected UDM Result To SSU Results"))
        self.save_all_ssu_figures_action.setText(self.tr("Save Figures For All SSU Results"))
        self.language_menu.setTitle(self.tr("Language"))
        self.theme_menu.setTitle(self.tr("Theme"))
        self.default_theme_action.setText(self.tr("Default"))
        self.light_theme_action.setText(self.tr("Light"))
        self.dark_theme_action.setText(self.tr("Dark"))
        self.log_action.setText(self.tr("Log"))
        self.about_action.setText(self.tr("About"))
        self.tab_widget.setTabText(0, self.tr("Generator"))
        self.tab_widget.setTabText(1, self.tr("Statistics"))
        self.tab_widget.setTabText(2, self.tr("PCA"))
        self.tab_widget.setTabText(3, self.tr("Clustering"))
        self.tab_widget.setTabText(4, self.tr("SSU"))
        self.tab_widget.setTabText(5, self.tr("EMMA"))
        self.tab_widget.setTabText(6, self.tr("UDM"))
        self.ssu_multicore_analyzer.retranslate()



__all__ = ["EMMASettings"]

import logging
from typing import *

from PySide6 import QtCore, QtWidgets
from grpc._channel import _InactiveRpcError

from ..protos.client import QGrainClient

built_in_losses = (
    "1-norm", "2-norm", "3-norm", "4-norm",
    "mae", "mse", "rmse", "rmlse", "lmse", "angular", "cosine")


class EMMASettings(QtWidgets.QDialog):
    logger = logging.getLogger("QGrain.EMMASettings")

    def __init__(self, client: QGrainClient = None, parent: QtWidgets.QWidget = None):
        super().__init__(parent=parent)
        self._client = client
        self.setWindowTitle(self.tr("EMMA Settings"))
        self.setAttribute(QtCore.Qt.WA_StyledBackground, True)
        self.main_layout = QtWidgets.QGridLayout(self)
        self.device_label = QtWidgets.QLabel(self.tr("Device"))
        self.device_label.setToolTip(self.tr("PyTorch can use the NVIDIA GPU to do calculations."))
        self.device_combo_box = QtWidgets.QComboBox()
        self.main_layout.addWidget(self.device_label, 0, 0)
        self.main_layout.addWidget(self.device_combo_box, 0, 1)
        self.loss_label = QtWidgets.QLabel(self.tr("Loss"))
        self.loss_label.setToolTip(self.tr(
            "The function to calculate the difference between prediction and observation."))
        self.loss_combo_box = QtWidgets.QComboBox()
        for key, name in self.supported_losses:
            self.loss_combo_box.addItem(name)
        self.loss_combo_box.setCurrentIndex(8)
        self.main_layout.addWidget(self.loss_label, 1, 0)
        self.main_layout.addWidget(self.loss_combo_box, 1, 1)
        self.pretrain_epochs_label = QtWidgets.QLabel(self.tr("Pretrain Epochs"))
        self.pretrain_epochs_label.setToolTip(self.tr(
            "The number of epochs before formal training. "
            "The pretrain process only update the proportions of end members."))
        self.pretrain_epochs_input = QtWidgets.QSpinBox()
        self.pretrain_epochs_input.setRange(0, 100000)
        self.pretrain_epochs_input.setValue(0)
        self.main_layout.addWidget(self.pretrain_epochs_label, 2, 0)
        self.main_layout.addWidget(self.pretrain_epochs_input, 2, 1)
        self.min_epochs_label = QtWidgets.QLabel(self.tr("Minimum Number of Epochs"))
        self.min_epochs_label.setToolTip(self.tr("Minimum number of epochs to be performed."))
        self.min_epochs_input = QtWidgets.QSpinBox()
        self.min_epochs_input.setRange(10, 10000)
        self.min_epochs_input.setValue(500)
        self.main_layout.addWidget(self.min_epochs_label, 3, 0)
        self.main_layout.addWidget(self.min_epochs_input, 3, 1)
        self.max_epochs_label = QtWidgets.QLabel(self.tr("Maximum Number of Epochs"))
        self.max_epochs_label.setToolTip(self.tr("Maximum number of epochs to be performed."))
        self.max_epochs_input = QtWidgets.QSpinBox()
        self.max_epochs_input.setRange(100, 100000)
        self.max_epochs_input.setValue(2000)
        self.main_layout.addWidget(self.max_epochs_label, 4, 0)
        self.main_layout.addWidget(self.max_epochs_input, 4, 1)
        self.precision_label = QtWidgets.QLabel(self.tr("Precision"))
        self.precision_label.setToolTip(self.tr("It controls the precision of final loss."))
        self.precision_input = QtWidgets.QDoubleSpinBox()
        self.precision_input.setRange(2.0, 20.0)
        self.precision_input.setValue(10.0)
        self.main_layout.addWidget(self.precision_label, 5, 0)
        self.main_layout.addWidget(self.precision_input, 5, 1)
        self.learning_rate_label = QtWidgets.QLabel(self.tr("Learning Rate (x10<sup>-3</sup>)"))
        self.learning_rate_label.setToolTip(self.tr(
            "The learning rate of the neural network to update its weights from gradient."))
        self.learning_rate_input = QtWidgets.QDoubleSpinBox()
        self.learning_rate_input.setDecimals(3)
        self.learning_rate_input.setRange(0.001, 1000)
        self.learning_rate_input.setValue(10)
        self.main_layout.addWidget(self.learning_rate_label, 6, 0)
        self.main_layout.addWidget(self.learning_rate_input, 6, 1)
        self.beta1_label = QtWidgets.QLabel(self.tr("Beta 1"))
        self.beta1_label.setToolTip(self.tr(
            "Betas are the coefficients used for computing running averages of gradient and its square."))
        self.beta1_input = QtWidgets.QDoubleSpinBox()
        self.beta1_input.setDecimals(4)
        self.beta1_input.setRange(0.0001, 0.9999)
        self.beta1_input.setValue(0.8000)
        self.main_layout.addWidget(self.beta1_label, 7, 0)
        self.main_layout.addWidget(self.beta1_input, 7, 1)
        self.beta2_label = QtWidgets.QLabel(self.tr("Beta 2"))
        self.beta2_label.setToolTip(self.tr(
            "Betas are the coefficients used for computing running averages of gradient and its square."))
        self.beta2_input = QtWidgets.QDoubleSpinBox()
        self.beta2_input.setDecimals(4)
        self.beta2_input.setRange(0.0001, 0.9999)
        self.beta2_input.setValue(0.5000)
        self.main_layout.addWidget(self.beta2_label, 8, 0)
        self.main_layout.addWidget(self.beta2_input, 8, 1)
        self.update_EMs_checkbox = QtWidgets.QCheckBox(self.tr("Update End Members"))
        self.update_EMs_checkbox.setToolTip(self.tr(
            "Update the distributions of end members while training the model or not."))
        self.update_EMs_checkbox.setChecked(True)
        self.main_layout.addWidget(self.update_EMs_checkbox, 9, 0, 1, 2)
        self.need_history_checkbox = QtWidgets.QCheckBox(self.tr("Need History"))
        self.need_history_checkbox.setToolTip(self.tr("Record the variation history of parameters or not."))
        self.need_history_checkbox.setChecked(True)
        self.main_layout.addWidget(self.need_history_checkbox, 10, 0, 1, 2)
        self._update_device_list()

    @property
    def supported_losses(self) -> Sequence[Tuple[str, str]]:
        losses = (("1-norm", self.tr("1 Norm")),
                  ("2-norm", self.tr("2 Norm")),
                  ("3-norm", self.tr("3 Norm")),
                  ("4-norm", self.tr("4 Norm")),
                  ("mae", self.tr("MAE")),
                  ("mse", self.tr("MSE")),
                  ("rmse", self.tr("RMSE")),
                  ("rmlse", self.tr("RMLSE")),
                  ("lmse", self.tr("LMSE")),
                  ("cosine", self.tr("Cosine")),
                  ("angular", self.tr("Angular")))
        return losses

    @property
    def loss(self) -> str:
        return self.supported_losses[self.loss_combo_box.currentIndex()][0]

    @property
    def settings(self) -> Dict[str, Any]:
        settings = dict(
            device=self.device_combo_box.currentText(),
            loss=self.loss,
            pretrain_epochs=self.pretrain_epochs_input.value(),
            min_epochs=self.min_epochs_input.value(),
            max_epochs=self.max_epochs_input.value(),
            precision=self.precision_input.value(),
            learning_rate=self.learning_rate_input.value() / 1000.0,
            betas=(self.beta1_input.value(), self.beta2_input.value()),
            update_end_members=self.update_EMs_checkbox.isChecked(),
            need_history=self.need_history_checkbox.isChecked())
        return settings

    @settings.setter
    def settings(self, s: Dict):
        loss_map = {key: i for i, (key, name) in enumerate(self.supported_losses)}
        assert s["loss"] in loss_map
        self.device_combo_box.setCurrentIndex(s["device"])
        self.loss_combo_box.setCurrentIndex(loss_map[s["loss"]])
        self.pretrain_epochs_input.setValue(s["pretrain_epochs"])
        self.min_epochs_input.setValue(s["min_epochs"])
        self.max_epochs_input.setValue(s["max_epochs"])
        self.precision_input.setValue(s["precision"])
        self.learning_rate_input.setValue(s["learning_rate"] * 1000.0)
        beta1, beta2 = s["betas"]
        self.beta1_input.setValue(beta1)
        self.beta2_input.setValue(beta2)
        self.update_EMs_checkbox.setChecked(s["update_end_members"])
        self.need_history_checkbox.setChecked(s["need_history"])

    def _update_device_list(self):
        if self._client is not None and self._client.has_target:
            try:
                server_state = self._client.get_service_state()
                available_devices = server_state["available_devices"]
                self.device_combo_box.addItems(available_devices)
                return
            except _InactiveRpcError:
                self.logger.warning("The remote grpc server is not available.")

        import torch
        available_devices = ["cpu"]
        if torch.cuda.is_available():
            available_devices.append("cuda")
        available_devices.extend([f"cuda:{i}" for i in range(torch.cuda.device_count())])
        if "cuda:0" in available_devices:
            available_devices.remove("cuda:0")
        self.device_combo_box.addItems(available_devices)

    def changeEvent(self, event: QtCore.QEvent):
        if event.type() == QtCore.QEvent.LanguageChange:
            self.retranslate()

    def retranslate(self):
        self.setWindowTitle(self.tr("EMMA Settings"))
        self.device_label.setText(self.tr("Device"))
        self.device_label.setToolTip(self.tr("PyTorch can use the NVIDIA GPU to do calculations."))
        self.loss_label.setText(self.tr("Loss"))
        self.loss_label.setToolTip(self.tr(
            "The function to calculate the difference between prediction and observation."))
        for i, (key, name) in enumerate(self.supported_losses):
            self.loss_combo_box.setItemText(i, name)
        self.pretrain_epochs_label.setText(self.tr("Pretrain Epochs"))
        self.pretrain_epochs_label.setToolTip(self.tr(
            "The number of epochs before formal training. "
            "The pretrain process only update the proportions of end members."))
        self.min_epochs_label.setText(self.tr("Minimum Number of Epochs"))
        self.min_epochs_label.setToolTip(self.tr("Minimum number of epochs to be performed."))
        self.max_epochs_label.setText(self.tr("Maximum Number of Epochs"))
        self.max_epochs_label.setToolTip(self.tr("Maximum number of epochs to be performed."))
        self.precision_label.setText(self.tr("Precision"))
        self.precision_label.setToolTip(self.tr("It controls the precision of final loss."))
        self.learning_rate_label.setText(self.tr("Learning Rate (x10<sup>-3</sup>)"))
        self.learning_rate_label.setToolTip(self.tr(
            "The learning rate of the neural network to update its weights from gradient."))
        self.beta1_label.setText(self.tr("Beta 1"))
        self.beta1_label.setToolTip(self.tr(
            "Betas are the coefficients used for computing running averages of gradient and its square."))
        self.beta2_label.setText(self.tr("Beta 2"))
        self.beta2_label.setToolTip(self.tr(
            "Betas are the coefficients used for computing running averages of gradient and its square."))
        self.update_EMs_checkbox.setText(self.tr("Update End Members"))
        self.update_EMs_checkbox.setToolTip(self.tr(
            "Update the distributions of end members while training the model or not."))
        self.need_history_checkbox.setText(self.tr("Need History"))
        self.need_history_checkbox.setToolTip(self.tr("Record the variation history of parameters or not."))



from typing import *

import grpc
import numpy as np
from numpy import ndarray

from ..models import DistributionType, KernelType, Sample, SSUResult, Dataset, ArtificialDataset, EMMAResult, UDMResult
from ..distributions import get_distribution
from . import qgrain_pb2 as qgrain_pb2
from . import qgrain_pb2_grpc as qgrain_pb2_grpc

MAX_MESSAGE_LENGTH = 2**30


class QGrainClient:
    _target = ""

    def __init__(self):
        pass

    @classmethod
    def set_target(cls, target: str):
        cls._target = target

    @classmethod
    def _to_sample_pb2(cls, sample: Union[ArtificialDataset, Sample]) -> qgrain_pb2.Sample:
        return qgrain_pb2.Sample(name=sample.name, classes=sample.classes, distribution=sample.distribution)

    @classmethod
    def _to_dataset_pb2(cls, dataset: Union[ArtificialDataset, Dataset]) -> qgrain_pb2.Dataset:
        dataset_pb2 = qgrain_pb2.Dataset(name=dataset.name, n_samples=len(dataset), n_classes=len(dataset.classes),
                                         sample_names=dataset.sample_names, classes=dataset.classes,
                                         distributions=dataset.distributions.astype(np.float64).tobytes())
        return dataset_pb2

    @classmethod
    def _statistical_result_pb2_to_dict(cls, response: qgrain_pb2.StatisticalRequest) -> Dict:
        arithmetic_keys = ("mean", "std", "skewness", "kurtosis")
        methods = ("geometric", "logarithmic", "geometric_fw57", "logarithmic_fw57")
        method_keys = ("mean", "std", "skewness", "kurtosis", "median", "mode", "modes",
                       "mean_description", "std_description", "skewness_description", "kurtosis_description")
        other_keys = ("proportions_gsm", "proportions_ssc", "proportions_bgssc", "proportions",
                      "group_folk54", "group_bp12_symbol", "group_bp12")
        statistics = {"arithmetic": {key: getattr(response.arithmetic, key) for key in arithmetic_keys}}
        for method in methods:
            statistics[method] = {key: getattr(getattr(response, method), key) for key in method_keys}
        for key in other_keys:
            statistics[key] = getattr(response, key)
        return statistics

    @property
    def has_target(self) -> bool:
        return self._target != ""

    def get_service_state(self):
        with grpc.insecure_channel(self._target, options=[
                ("grpc.max_send_message_length", MAX_MESSAGE_LENGTH),
                ("grpc.max_receive_message_length", MAX_MESSAGE_LENGTH)]) as channel:
            stub = qgrain_pb2_grpc.QGrainStub(channel)
            request = qgrain_pb2.ServiceStateRequest()
            response = stub.get_service_state(request)
            state = dict(max_workers=response.max_workers, max_message_length=response.max_message_length,
                         available_devices=tuple(response.available_devices),
                         max_dataset_size=response.max_dataset_size)
            return state

    def get_statistics(self, dataset: Union[ArtificialDataset, Dataset]) -> Optional[List[Dict[str, Any]]]:
        with grpc.insecure_channel(self._target, options=[
                ('grpc.max_send_message_length', MAX_MESSAGE_LENGTH),
                ('grpc.max_receive_message_length', MAX_MESSAGE_LENGTH)]) as channel:
            stub = qgrain_pb2_grpc.QGrainStub(channel)
            request = qgrain_pb2.StatisticalRequest(dataset=self._to_dataset_pb2(dataset))
            response = stub.get_statistics(request)
            if len(response.results) != 0:
                results = [self._statistical_result_pb2_to_dict(result) for result in response.results]
                return results
            else:
                return response.message

    def get_ssu_result(self, sample: Union[ArtificialDataset, Sample], distribution_type: DistributionType,
                       n_components: int, x0: ndarray = None, loss: str = "lmse", optimizer: str = "L-BFGS-B",
                       try_global: bool = False, global_max_niter: int = 100, global_niter_success: int = 5,
                       global_step_size: float = 0.2, optimizer_max_niter: int = 10000, need_history: bool = True) -> \
            Union[SSUResult, str]:
        with grpc.insecure_channel(self._target, options=[
                ('grpc.max_send_message_length', MAX_MESSAGE_LENGTH),
                ('grpc.max_receive_message_length', MAX_MESSAGE_LENGTH)]) as channel:
            stub = qgrain_pb2_grpc.QGrainStub(channel)
            sample_pb2 = self._to_sample_pb2(sample)
            distribution_type_pb2 = distribution_type.value
            x0_pb2 = bytes() if x0 is None else x0.astype(np.float64).tobytes()
            settings = dict(loss=loss, optimizer=optimizer, try_global=try_global, global_max_niter=global_max_niter,
                            global_niter_success=global_niter_success, global_step_size=global_step_size,
                            optimizer_max_niter=optimizer_max_niter, need_history=need_history)
            request = qgrain_pb2.SSURequest(sample=sample_pb2, distribution_type=distribution_type_pb2,
                                            n_components=n_components, x0=x0_pb2, **settings)
            response = stub.get_ssu_result(request)
            # if not success, the `parameters` is empty bytes
            if len(response.parameters) != 0:
                shape = (response.n_iterations, response.n_parameters, response.n_components)
                parameters = np.frombuffer(response.parameters, np.float64).copy().reshape(shape)
                result = SSUResult(sample, distribution_type, parameters, response.time_spent, x0=x0, settings=settings)
                return result
            else:
                return response.message

    def get_emma_result(self, dataset: Union[ArtificialDataset, Dataset], kernel_type: KernelType, n_members: int,
                        x0: ndarray = None, device="cpu", loss="lmse", pretrain_epochs=0, min_epochs=100,
                        max_epochs=10000, precision: Union[int, float] = 6, learning_rate=5e-3, betas=(0.8, 0.5),
                        update_end_members=True, need_history=True) -> Union[EMMAResult, str]:
        with grpc.insecure_channel(self._target, options=[
                ('grpc.max_send_message_length', MAX_MESSAGE_LENGTH),
                ('grpc.max_receive_message_length', MAX_MESSAGE_LENGTH)]) as channel:
            stub = qgrain_pb2_grpc.QGrainStub(channel)
            dataset_pb2 = self._to_dataset_pb2(dataset)
            x0_pb2 = bytes() if x0 is None else x0.astype(np.float64).tobytes()
            settings = dict(device=device, loss=loss, pretrain_epochs=pretrain_epochs, min_epochs=min_epochs,
                            max_epochs=max_epochs, precision=precision, learning_rate=learning_rate, betas=betas,
                            update_end_members=update_end_members, need_history=need_history)
            request = qgrain_pb2.EMMARequest(dataset=dataset_pb2, distribution_type=kernel_type.value,
                                             n_members=n_members, x0=x0_pb2, **settings)
            response = stub.get_emma_result(request)
            if len(response.losses) != 0:
                proportions_shape = (response.n_iterations, response.n_samples, response.n_members)
                end_members_shape = (response.n_iterations, response.n_members, response.n_classes)
                proportions = np.frombuffer(response.proportions, dtype=np.float64).copy().reshape(proportions_shape)
                end_members = np.frombuffer(response.end_members, dtype=np.float64).copy().reshape(end_members_shape)
                losses = {loss: np.array(response.losses, dtype=np.float64)}
                result = EMMAResult(dataset, kernel_type, n_members, proportions, end_members, response.time_spent,
                                    x0=x0, loss_series=losses, settings=settings)
                return result
            else:
                return response.message

    def get_udm_result(self, dataset: Union[ArtificialDataset, Dataset], kernel_type: KernelType, n_components: int,
                       x0: ndarray = None, device="cpu", pretrain_epochs=0, min_epochs=100, max_epochs=10000,
                       precision: Union[int, float] = 6, learning_rate=5e-3, betas=(0.8, 0.5),
                       constraint_level: Union[int, float] = 2.0, need_history=True) -> Union[UDMResult, str]:
        with grpc.insecure_channel(self._target, options=[
                ('grpc.max_send_message_length', MAX_MESSAGE_LENGTH),
                ('grpc.max_receive_message_length', MAX_MESSAGE_LENGTH)]) as channel:
            stub = qgrain_pb2_grpc.QGrainStub(channel)
            dataset_pb2 = self._to_dataset_pb2(dataset)
            x0_pb2 = bytes() if x0 is None else x0.astype(np.float64).tobytes()
            n_parameters = get_distribution(DistributionType.__members__[kernel_type.name]).N_PARAMETERS + 1
            settings = dict(device=device, pretrain_epochs=pretrain_epochs, min_epochs=min_epochs,
                            max_epochs=max_epochs, precision=precision, learning_rate=learning_rate, betas=betas,
                            constraint_level=constraint_level, need_history=need_history)
            request = qgrain_pb2.UDMRequest(dataset=dataset_pb2, distribution_type=kernel_type.value,
                                            n_components=n_components, x0=x0_pb2, **settings)
            response = stub.get_udm_result(request)
            if len(response.parameters) != 0:
                shape = (response.n_iterations, response.n_samples, n_parameters, response.n_components)
                parameters = np.frombuffer(response.parameters, dtype=np.float64).copy().reshape(shape)
                losses = {"total": np.array(response.total_losses, dtype=np.float64),
                          "distribution": np.array(response.distribution_losses, dtype=np.float64),
                          "component": np.array(response.component_losses, dtype=np.float64)}
                result = UDMResult(dataset, kernel_type, n_components, parameters, response.time_spent, x0=x0,
                                   loss_series=losses, settings=settings)
                return result
            else:
                return response.message



from typing import *

from numpy import ndarray
from scipy.cluster.hierarchy import dendrogram, linkage
from sklearn.decomposition import PCA

from . import BaseChart
from ..models import ArtificialDataset, Dataset


class HierarchicalChart(BaseChart):
    def __init__(self, parent=None, figsize=(6.6, 4.4)):
        super().__init__(parent=parent, figsize=figsize)
        self.setWindowTitle(self.tr("Hierarchical Clustering"))
        self._axes = self._figure.subplots()
        self._last_result = None

    def show_dataset(self, dataset: Union[ArtificialDataset, Dataset], method="ward", metric="euclidean", p=100):
        pca = PCA(n_components=0.95)
        transformed = pca.fit_transform(dataset.distributions)
        linkage_matrix = linkage(transformed, method=method, metric=metric)
        self.show_matrix(linkage_matrix, p)

    def show_matrix(self, linkage_matrix: ndarray, p=100):
        self._last_result = (linkage_matrix, p)
        self._axes.clear()
        dendrogram(linkage_matrix, p=p, truncate_mode="lastp", no_labels=False, leaf_font_size=7, ax=self._axes)
        self._axes.set_xlabel("Sample count/index")
        self._axes.set_ylabel("Distance")
        self._figure.tight_layout()
        self._canvas.draw()

    def update_chart(self):
        self._figure.clear()
        self._axes = self._figure.subplots()
        if self._last_result is not None:
            self.show_matrix(*self._last_result)

    def retranslate(self):
        self.setWindowTitle(self.tr("Hierarchical Clustering"))
        self.edit_figure_action.setText(self.tr("Edit Figure"))
        self.save_figure_action.setText(self.tr("Save Figure"))



__all__ = ["ArtificialComponent", "ArtificialSample", "ArtificialDataset"]

from typing import *

import numpy as np
from numpy import ndarray

from ..statistics import to_phi, to_microns, interval_phi
from ..models import DistributionType, Dataset, Sample
from ..distributions import get_distribution


class ArtificialComponent:
    """The class to represent one component of the artificial sample."""
    __slots__ = ("_classes", "_classes_phi", "_distribution", "_proportion", "_moments")

    def __init__(self, classes: ndarray, classes_phi: ndarray,
                 distribution: ndarray, proportion: float,
                 moments: Tuple[float, float, float, float]):
        """
        Construct an instance of the `ArtificialComponent` class.

        **If not necessary, do not manually create the sample, because it will not validate the passed parameters.**

        :param classes: The grain size classes in microns.
        :param classes_phi: The grain size classes in phi values.
        :param distribution: The frequency distribution of grain size classes.
            Note, the sum of frequencies should be equal to 1.
        :param moments: A tuple that contains the mean, standard deviation, skewness, and kurtosis, respectively.
        :return: An instance of the `ArtificialComponent` class.
        """
        self._classes = classes
        self._classes_phi = classes_phi
        self._distribution = distribution
        self._proportion = proportion
        m, std, s, k = moments
        self._moments = dict(mean=m, std=std, skewness=s, kurtosis=k)

    def __repr__(self):
        return f"AC({self._moments['mean']:.2f} phi, {self._proportion:.2%})"

    @property
    def classes(self) -> ndarray:
        return self._classes

    @property
    def classes_phi(self) -> ndarray:
        return self._classes_phi

    @property
    def interval_phi(self) -> float:
        return interval_phi(self.classes_phi)

    @property
    def distribution(self) -> ndarray:
        return self._distribution

    @property
    def proportion(self) -> float:
        return self._proportion

    @property
    def moments(self) -> Dict[str, float]:
        return self._moments

    @property
    def mean(self) -> float:
        return self._moments["mean"]

    @property
    def sorting_coefficient(self) -> float:
        return self._moments["std"]

    @property
    def skewness(self) -> float:
        return self._moments["skewness"]

    @property
    def kurtosis(self) -> float:
        return self._moments["kurtosis"]


class ArtificialSample:
    """The class to represent one sample of the artificial dataset."""
    __slots__ = ("_name", "_classes", "_classes_phi", "_distribution", "_components", "_proportions", "_moments")

    def __init__(self, name: str, classes: ndarray, classes_phi: ndarray,
                 distribution: ndarray, components: Sequence[ndarray], proportions: Sequence[float],
                 moments: Tuple[Sequence, Sequence, Sequence, Sequence]):
        """
        Construct an instance of the `ArtificialSample` class.

        **If not necessary, do not manually create the sample, because it will not validate the passed parameters.**

        :param name: The name of this artificial sample. It can be modified.
        :param classes: The grain size classes in microns.
        :param classes_phi: The grain size classes in phi values.
        :param distribution: The frequency distribution of grain size classes.
            Note, the sum of frequencies should be equal to 1.
        :param components: The grain size distributions of all components.
        :param proportions: The proportions of all components.
        :param moments: A tuple that contains the mean, standard deviation, skewness,
            and kurtosis of all components, respectively.
        :return: An instance of the `ArtificialSample` class.
        """
        self._name = name
        self._classes = classes
        self._classes_phi = classes_phi
        self._distribution = distribution
        self._components = components
        self._proportions = proportions
        self._moments = moments

    def __repr__(self):
        return f"AS({self._name}, {len(self._components)})"

    def __len__(self):
        return len(self._components)

    def __iter__(self):
        for i in range(len(self._components)):
            yield self._get_component(i)

    def __getitem__(self, item):
        if isinstance(item, int):
            return self._get_component(item)
        elif isinstance(item, slice):
            return [self._get_component(index) for index in np.arange(len(self._components))[item]]
        else:
            raise TypeError(f"Component indices must be integers or slices, not {type(item)}.")

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str):
        assert isinstance(value, str)
        assert len(value) > 0
        self._name = value

    @property
    def classes(self) -> ndarray:
        return self._classes

    @property
    def classes_phi(self) -> ndarray:
        return self._classes_phi

    @property
    def interval_phi(self) -> float:
        return interval_phi(self._classes_phi)

    @property
    def distribution(self) -> ndarray:
        return self._distribution

    @property
    def sample(self):
        sample = Sample(self._name, self._classes, self._classes_phi, self._distribution)
        return sample

    @property
    def is_valid(self) -> bool:
        valid = True
        for values in [self._proportions, self._components, self._distribution, *self._moments]:
            if np.any(np.logical_or(np.isnan(values), np.isinf(values))):
                valid = False
                break
        return valid

    def _get_component(self, index: int):
        m, std, s, k = self._moments
        component = ArtificialComponent(
            self._classes, self._classes_phi,
            self._components[index], self._proportions[index],
            (m[index], std[index], s[index], k[index]))
        return component


class ArtificialDataset:
    """
    The class to represent one artificial dataset.

    * Get the sample at index i, `sample = dataset[i]`.
    * Iterate all samples, `for sample in dataset`.
    * Iterate partial samples, `for sample in dataset[:10]`.
    * Get the number of samples, `len(dataset)`.
    """
    def __init__(self, parameters: ndarray,
                 distribution_type: DistributionType,
                 min_size=0.02, max_size=2000.0, n_classes=101,
                 precision=4, noise=5):
        """
        Construct an instance of the `ArtificialDataset` class.

        :param parameters: A three-dimensional numpy array that contains the parameters to generate samples.
        :param distribution_type: The type of elementary distribution.
        :param min_size: The minimum grain size class in microns.
        :param max_size: The maximum grain size class in microns.
        :param n_classes: The number of grain size classes. More classes can make the distributions more detailed.
        :param precision: The precision level (decimals) of generated grain size distributions.
        :param noise: The standard deviation of the random noise matrix. Usually, should use `precision+1`.
        :return: An instance of the `ArtificialDataset` class.
        """
        # validations
        assert isinstance(parameters, ndarray)
        assert parameters.ndim == 3
        assert isinstance(min_size, (int, float))
        assert isinstance(max_size, (int, float))
        assert isinstance(n_classes, int)
        assert isinstance(precision, int)
        assert isinstance(noise, int)
        assert min_size > 0
        assert max_size > 0
        assert max_size > min_size
        assert n_classes > 1
        assert precision > 1
        assert noise > 1
        n_samples, n_parameters, n_components = parameters.shape
        distribution_class = get_distribution(distribution_type)
        assert n_parameters == distribution_class.N_PARAMETERS + 1
        assert n_samples > 0
        assert n_parameters == 3 or n_parameters == 4
        assert n_components > 0
        # preparation
        self._name = f"AD({n_samples}, {n_components}, {distribution_type.name})"
        self._parameters = parameters
        self._distribution_type = distribution_type
        self._min_size, self._max_size = min_size, max_size
        self._classes_phi = np.linspace(to_phi(min_size), to_phi(max_size), n_classes)
        self._classes = to_microns(self._classes_phi)
        self._precision = precision
        self._noise = noise
        classes = np.expand_dims(np.expand_dims(self._classes_phi, 0), 0).repeat(n_samples, 0).repeat(n_components, 1)
        proportions, components, (m, std, s, k) = distribution_class.interpret(
            parameters, classes, interval_phi(self._classes_phi))
        noise = np.random.randn(n_samples, n_classes) * (10 ** (-noise))
        distributions = np.round((proportions @ components).squeeze(1) + noise, decimals=precision)
        self._proportions = proportions
        self._components = components
        self._distributions = distributions
        self._mean = m
        self._std = std
        self._skewness = s
        self._kurtosis = k

    def __repr__(self):
        return f"AD({self._parameters.shape[0]}, {self._parameters.shape[2]}, {self._distribution_type.name})"

    def __len__(self):
        return self._parameters.shape[0]

    def __iter__(self):
        for i in range(self._parameters.shape[0]):
            yield self._get_sample(i)

    def __getitem__(self, item):
        if isinstance(item, int):
            return self._get_sample(item)
        elif isinstance(item, slice):
            return [self._get_sample(index) for index in np.arange(self._parameters.shape[0])[item]]
        else:
            raise TypeError(f"Sample indices must be integers or slices, not {type(item)}.")

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str):
        assert isinstance(value, str)
        assert len(value) > 0
        self._name = value

    @property
    def sample_names(self) -> List[str]:
        return [f"AS{i+1}" for i in range(self._parameters.shape[0])]

    @property
    def parameters(self) -> ndarray:
        return self._parameters

    @property
    def distribution_type(self) -> DistributionType:
        return self._distribution_type

    @property
    def n_samples(self) -> int:
        return self._parameters.shape[0]

    @property
    def n_parameters(self) -> int:
        return self._parameters.shape[1]

    @property
    def n_components(self) -> int:
        return self._parameters.shape[2]

    @property
    def n_classes(self) -> int:
        return len(self._classes)

    @property
    def classes(self) -> ndarray:
        return self._classes

    @property
    def classes_phi(self) -> ndarray:
        return self._classes_phi

    @property
    def interval_phi(self) -> float:
        return interval_phi(self._classes_phi)

    @property
    def noise(self) -> int:
        return self._noise

    @property
    def precision(self) -> int:
        return self._precision

    @property
    def distributions(self) -> ndarray:
        return self._distributions

    @property
    def components(self) -> ndarray:
        return self._components

    @property
    def proportions(self) -> ndarray:
        return self._proportions

    @property
    def dataset(self) -> Dataset:
        sample_names = [f"AS{i + 1}" for i in range(self.n_samples)]
        dataset = Dataset("Artificial Dataset", sample_names, self._classes, self._distributions)
        return dataset

    def _get_sample(self, index: int) -> ArtificialSample:
        name = f"AS{index + 1}"
        distribution = self._distributions[index]
        proportions = self._proportions[index, 0, :]
        components = self._components[index]
        moments = self._mean[index], self._std[index], self._skewness[index], self._kurtosis[index]
        sample = ArtificialSample(
            name, self._classes, self._classes_phi,
            distribution, components, proportions, moments)
        return sample



__all__ = ["get_image_by_proportions", "udm_to_ssu"]

import logging
from typing import *

import numpy as np
from numpy import ndarray

from QGrain.models import DistributionType, UDMResult, SSUResult


def get_image_by_proportions(proportions: ndarray, resolution: int = 100) -> ndarray:
    n_samples, n_components = proportions.shape
    index = np.repeat(np.expand_dims(np.linspace(0.0, 1.0, resolution), axis=0), n_samples, axis=0)
    image = np.zeros((n_samples, resolution))

    bound = np.ones((n_samples, n_components+1))
    bound[:, 0] = 0.0
    bottom = np.zeros(n_samples)
    for i in range(n_components):
        bottom += proportions[:, i]
        bound[:, i+1] = bottom
    bound[:, -1] = 1.0

    for i in range(n_components):
        lower = np.repeat(np.expand_dims(bound[:, i], axis=1), resolution, axis=1)
        upper = np.repeat(np.expand_dims(bound[:, i+1], axis=1), resolution, axis=1)
        key = np.logical_and(np.greater_equal(index, lower), np.less_equal(index, upper))
        image[key] = i
    return image.T


def udm_to_ssu(result: UDMResult, logger: logging.Logger = None,
               progress_callback: Callable[[float], None] = None) -> List[SSUResult]:
    assert isinstance(result, UDMResult)
    if logger is None:
        logger = logging.getLogger("QGrain")
    else:
        assert isinstance(logger, logging.Logger)
    distribution_type = DistributionType.__members__[result.kernel_type.name]
    weight = np.ones((1, result.n_components))
    x0 = np.concatenate([result.x0, weight], axis=0).astype(np.float64)
    time_spent = result.time_spent / result.n_samples
    ssu_results = []
    for i in range(result.n_samples):
        if result.n_iterations == 0:
            parameters = np.expand_dims(result.parameters[-1, i], axis=0)
        else:
            history = [np.expand_dims(result.parameters[j][i], axis=0) for j in range(result.n_iterations)]
            parameters = np.concatenate(history, axis=0)
        ssu_result = SSUResult(result.dataset[i], distribution_type, parameters, time_spent,
                               x0=x0, settings=result.settings)
        ssu_results.append(ssu_result)
        if progress_callback is not None:
            progress_callback(i / result.n_samples)
    if progress_callback is not None:
        progress_callback(1.0)
    return ssu_results



__all__ = ["About"]

import os

from PySide6 import QtCore, QtWidgets

from . import QGRAIN_ROOT_PATH


class About(QtWidgets.QDialog):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.setWindowTitle(self.tr("About"))
        self.setMinimumSize(400, 400)
        self.main_layout = QtWidgets.QGridLayout(self)
        self.text = QtWidgets.QTextBrowser()
        self.text.setReadOnly(True)
        self.text.setOpenExternalLinks(True)
        self.main_layout.addWidget(self.text, 0, 0)
        with open(os.path.join(QGRAIN_ROOT_PATH, "assets", "README.md"), "r") as f:
            md = f.read()
            self.text.setMarkdown(md)

    def changeEvent(self, event: QtCore.QEvent):
        if event.type() == QtCore.QEvent.LanguageChange:
            self.retranslate()

    def retranslate(self):
        self.setWindowTitle(self.tr("About"))



import logging
import os
from typing import *

import matplotlib.pyplot as plt
from PySide6 import QtCore, QtGui, QtWidgets
from matplotlib.animation import FFMpegWriter, FuncAnimation, ImageMagickWriter
from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qtagg import NavigationToolbar2QT as NavigationToolbar


def normal_color():
    return os.environ["QTMATERIAL_SECONDARYTEXTCOLOR"]


def highlight_color():
    return os.environ["QTMATERIAL_PRIMARYCOLOR"]


def background_color():
    return os.environ["QTMATERIAL_SECONDARYDARKCOLOR"]


def synchronize_theme():
    plt.rcParams["axes.facecolor"] = os.environ["QTMATERIAL_SECONDARYDARKCOLOR"]
    plt.rcParams["figure.facecolor"] = os.environ["QTMATERIAL_SECONDARYDARKCOLOR"]
    plt.rcParams["axes.edgecolor"] = os.environ["QTMATERIAL_SECONDARYTEXTCOLOR"]
    plt.rcParams["axes.titlecolor"] = os.environ["QTMATERIAL_SECONDARYTEXTCOLOR"]
    plt.rcParams["axes.labelcolor"] = os.environ["QTMATERIAL_SECONDARYTEXTCOLOR"]
    plt.rcParams["grid.color"] = os.environ["QTMATERIAL_SECONDARYTEXTCOLOR"]
    plt.rcParams["legend.labelcolor"] = os.environ["QTMATERIAL_SECONDARYTEXTCOLOR"]
    plt.rcParams["xtick.color"] = os.environ["QTMATERIAL_SECONDARYTEXTCOLOR"]
    plt.rcParams["xtick.labelcolor"] = os.environ["QTMATERIAL_SECONDARYTEXTCOLOR"]
    plt.rcParams["ytick.color"] = os.environ["QTMATERIAL_SECONDARYTEXTCOLOR"]
    plt.rcParams["ytick.labelcolor"] = os.environ["QTMATERIAL_SECONDARYTEXTCOLOR"]

    plt.rcParams["savefig.dpi"] = 300.0
    plt.rcParams["savefig.transparent"] = False
    plt.rcParams["figure.max_open_warning"] = False


def setup_matplotlib():
    plt.style.use(["science", "no-latex"])
    plt.set_cmap("tab10")
    plt.rcParams['axes.unicode_minus'] = False
    plt.rcParams["font.family"] = "DejaVu Sans"
    plt.rcParams["font.size"] = 8
    plt.rcParams["axes.titlesize"] = 8
    plt.rcParams["axes.labelsize"] = 8
    plt.rcParams["xtick.labelsize"] = 7
    plt.rcParams["ytick.labelsize"] = 7
    plt.rcParams["legend.title_fontsize"] = 8
    plt.rcParams["legend.fontsize"] = 7
    plt.rcParams["mathtext.fontset"] = "dejavusans"
    synchronize_theme()


class BaseChart(QtWidgets.QWidget):
    logger = logging.getLogger("QGrain.charts")

    def __init__(self, parent=None, figsize=(4, 3)):
        super().__init__(parent=parent)
        self._figure: plt.Figure = plt.figure(figsize=figsize)
        self._canvas = FigureCanvas(self._figure)
        self._toolbar = NavigationToolbar(self._canvas, self)
        self.main_layout = QtWidgets.QGridLayout(self)
        self.main_layout.setContentsMargins(0, 0, 0, 0)
        self.main_layout.addWidget(self._canvas, 0, 0)
        self.menu = QtWidgets.QMenu(self._canvas)
        self.menu.setShortcutAutoRepeat(True)
        self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.show_menu)
        self.edit_figure_action = self.menu.addAction(self.tr("Edit Figure"))
        self.edit_figure_action.triggered.connect(lambda: self._toolbar.edit_parameters())
        self.save_figure_action = self.menu.addAction(self.tr("Save Figure"))
        self.save_figure_action.triggered.connect(lambda: self._toolbar.save_figure())
        self.normal_msg = QtWidgets.QMessageBox(parent=self)
        self._animation: Optional[FuncAnimation] = None

    def show_message(self, title: str, message: str):
        self.normal_msg.setWindowTitle(title)
        self.normal_msg.setText(message)
        self.normal_msg.exec_()

    def show_info(self, message: str):
        self.show_message(self.tr("Info"), message)

    def show_warning(self, message: str):
        self.show_message(self.tr("Warning"), message)

    def show_error(self, message: str):
        self.show_message(self.tr("Error"), message)

    def show_menu(self, pos: QtCore.QPoint):
        self.menu.popup(QtGui.QCursor.pos())

    def update_chart(self):
        pass

    def save_chart(self, filename: str, **kwargs):
        self._figure.savefig(filename, **kwargs)

    def save_animation(self, filename: str = None):
        if self._animation is None:
            return
        self._animation.pause()
        if filename is None:
            filename, format_str = QtWidgets.QFileDialog.getSaveFileName(
                self, self.tr("Choose a filename to save the animation of this SSU result"),
                ".", "MPEG-4 Video File (*.mp4);;Html Animation (*.html);;Graphics Interchange Format (*.gif)")
        if filename is None or filename == "":
            return
        progress_dialog = QtWidgets.QProgressDialog(
            self.tr("Saving Animation Frames..."), self.tr("Cancel"),
            0, 100, self)
        progress_dialog.setWindowTitle("QGrain")
        progress_dialog.setWindowModality(QtCore.Qt.WindowModal)

        def callback(frame_number, total_frames):
            if progress_dialog.wasCanceled():
                raise StopIteration()
            progress_dialog.setValue(int(frame_number / total_frames * 100))
            QtCore.QCoreApplication.processEvents()

        try:
            if filename[-5:] == ".html":
                if not FFMpegWriter.isAvailable():
                    self.show_error(self.tr("FFMpeg is not installed."))
                else:
                    self.show_info(self.tr("Rendering the animation to a html5 video, it will take several minutes."))
                    html = self._animation.to_html5_video()
                    with open(filename, "w") as f:
                        f.write(html)
            elif filename[-4:] == ".gif":
                if not ImageMagickWriter.isAvailable():
                    self.show_error(self.tr("ImageMagick is not installed."))
                else:
                    self._animation.save(filename, writer="imagemagick", fps=10, progress_callback=callback)
            elif filename[-4:] == ".mp4":
                if not FFMpegWriter.isAvailable():
                    self.show_error(self.tr("FFMpeg is not installed."))
                else:
                    self._animation.save(filename, writer="ffmpeg", fps=10, progress_callback=callback)
        except StopIteration:
            self.logger.info("The saving task was canceled.")
        finally:
            progress_dialog.close()

    def retranslate(self):
        pass

    def changeEvent(self, event: QtCore.QEvent):
        if event.type() == QtCore.QEvent.StyleChange:
            setup_matplotlib()
            self._figure.clear()
            self.main_layout.removeWidget(self._canvas)
            self._canvas.setVisible(False)
            self._figure = plt.figure(figsize=self._figure.get_size_inches())
            self._canvas = FigureCanvas(self._figure)
            self._toolbar = NavigationToolbar(self._canvas, self)
            self.main_layout.addWidget(self._canvas, 0, 0)
            self.update_chart()
        elif event.type() == QtCore.QEvent.LanguageChange:
            self.retranslate()


from .BoxplotChart import BoxplotChart
from .FrequencyChart import FrequencyChart
from .Frequency3DChart import Frequency3DChart
from .FrequencyHeatmap import FrequencyHeatmap
from .FrequencyGroupChart import FrequencyGroupChart
from .CumulativeChart import CumulativeChart
from .diagrams import *
from .HierarchicalChart import HierarchicalChart
from .PCAResultChart import PCAResultChart
from .LossSeriesChart import LossSeriesChart
from .DistributionChart import DistributionChart
from .EMMAResultChart import EMMAResultChart
from .UDMResultChart import UDMResultChart



from typing import *

import matplotlib.pyplot as plt
import numpy as np
from sklearn.decomposition import PCA

from . import BaseChart
from . import normal_color
from ..models import Dataset, ArtificialDataset


class PCAResultChart(BaseChart):
    def __init__(self, parent=None, size=(6.6, 4.4)):
        super().__init__(parent=parent, figsize=size)
        self.sample_axes = self._figure.add_subplot(2, 2, 1)
        self.shape_axes = self._figure.add_subplot(2, 2, 2)
        self.series_axes = self._figure.add_subplot(2, 1, 2)
        self.setWindowTitle(self.tr("PCA Chart"))
        self._last_dataset: Optional[Dataset] = None

    def show_dataset(self, dataset: Union[ArtificialDataset, Dataset]):
        assert dataset is not None
        self._last_dataset = dataset
        pca = PCA()
        transformed = pca.fit_transform(dataset.distributions)
        self.sample_axes.clear()
        self.shape_axes.clear()
        self.series_axes.clear()
        cmap = plt.get_cmap()
        self.sample_axes.scatter(transformed[:, 0], transformed[:, 1], c=normal_color(), s=2.0, alpha=0.05)
        self.sample_axes.plot(pca.components_[0, :], pca.components_[1, :], color=normal_color(), lw=1.0)
        xi = np.argmax(pca.components_[0, :])
        yi = np.argmax(pca.components_[1, :])
        self.sample_axes.arrow(0, 0, pca.components_[0, xi], pca.components_[1, xi], color=cmap(0), width=0.001)
        self.sample_axes.text(pca.components_[0, xi], pca.components_[1, xi], f"{dataset.classes[xi]: 0.4f}",
                              color=cmap(0), ha='center', va='center')
        self.sample_axes.arrow(0, 0, pca.components_[0, yi], pca.components_[1, yi], color=cmap(1), width=0.001)
        self.sample_axes.text(pca.components_[0, yi], pca.components_[1, yi], f"{dataset.classes[yi]: 0.4f}",
                              color=cmap(1), ha='center', va='center')
        self.sample_axes.set_xlabel("PC1")
        self.sample_axes.set_ylabel("PC2")
        cumulative_ratio = 0.0
        for i, ratio in enumerate(pca.explained_variance_ratio_):
            cumulative_ratio += ratio
            self.shape_axes.plot(dataset.classes, pca.components_[i], color=cmap(i), label=f"PC{i+1}")
            self.series_axes.plot(transformed[:, i], color=cmap(i),
                                  label=f"PC{i + 1} ({pca.explained_variance_ratio_[i]:0.2%})",
                                  lw=1.0, alpha=0.8)
            if i > 0 and cumulative_ratio > 0.95:
                break
        self.shape_axes.set_xscale("log")
        self.shape_axes.set_xlabel("Grain size (microns)")
        self.shape_axes.set_ylabel("Transformed value")
        self.shape_axes.legend(loc="upper left")
        self.series_axes.set_xlabel("Sample index")
        self.series_axes.set_ylabel("Transformed value")
        self.series_axes.legend(loc="upper left")
        self._figure.tight_layout()
        self._canvas.draw()

    def update_chart(self):
        self._figure.clear()
        self.sample_axes = self._figure.add_subplot(2, 2, 1)
        self.shape_axes = self._figure.add_subplot(2, 2, 2)
        self.series_axes = self._figure.add_subplot(2, 1, 2)
        if self._last_dataset is not None:
            self.show_dataset(self._last_dataset)

    def retranslate(self):
        self.setWindowTitle(self.tr("PCA Chart"))
        self.edit_figure_action.setText(self.tr("Edit Figure"))
        self.save_figure_action.setText(self.tr("Save Figure"))
